"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2025 LiveKit, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _DataChannelState:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _DataChannelStateEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_DataChannelState.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    DC_CONNECTING: _DataChannelState.ValueType  # 0
    DC_OPEN: _DataChannelState.ValueType  # 1
    DC_CLOSING: _DataChannelState.ValueType  # 2
    DC_CLOSED: _DataChannelState.ValueType  # 3

class DataChannelState(_DataChannelState, metaclass=_DataChannelStateEnumTypeWrapper): ...

DC_CONNECTING: DataChannelState.ValueType  # 0
DC_OPEN: DataChannelState.ValueType  # 1
DC_CLOSING: DataChannelState.ValueType  # 2
DC_CLOSED: DataChannelState.ValueType  # 3
global___DataChannelState = DataChannelState

class _QualityLimitationReason:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _QualityLimitationReasonEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_QualityLimitationReason.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    LIMITATION_NONE: _QualityLimitationReason.ValueType  # 0
    LIMITATION_CPU: _QualityLimitationReason.ValueType  # 1
    LIMITATION_BANDWIDTH: _QualityLimitationReason.ValueType  # 2
    LIMITATION_OTHER: _QualityLimitationReason.ValueType  # 3

class QualityLimitationReason(_QualityLimitationReason, metaclass=_QualityLimitationReasonEnumTypeWrapper): ...

LIMITATION_NONE: QualityLimitationReason.ValueType  # 0
LIMITATION_CPU: QualityLimitationReason.ValueType  # 1
LIMITATION_BANDWIDTH: QualityLimitationReason.ValueType  # 2
LIMITATION_OTHER: QualityLimitationReason.ValueType  # 3
global___QualityLimitationReason = QualityLimitationReason

class _IceRole:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _IceRoleEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_IceRole.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    ICE_UNKNOWN: _IceRole.ValueType  # 0
    ICE_CONTROLLING: _IceRole.ValueType  # 1
    ICE_CONTROLLED: _IceRole.ValueType  # 2

class IceRole(_IceRole, metaclass=_IceRoleEnumTypeWrapper): ...

ICE_UNKNOWN: IceRole.ValueType  # 0
ICE_CONTROLLING: IceRole.ValueType  # 1
ICE_CONTROLLED: IceRole.ValueType  # 2
global___IceRole = IceRole

class _DtlsTransportState:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _DtlsTransportStateEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_DtlsTransportState.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    DTLS_TRANSPORT_NEW: _DtlsTransportState.ValueType  # 0
    DTLS_TRANSPORT_CONNECTING: _DtlsTransportState.ValueType  # 1
    DTLS_TRANSPORT_CONNECTED: _DtlsTransportState.ValueType  # 2
    DTLS_TRANSPORT_CLOSED: _DtlsTransportState.ValueType  # 3
    DTLS_TRANSPORT_FAILED: _DtlsTransportState.ValueType  # 4

class DtlsTransportState(_DtlsTransportState, metaclass=_DtlsTransportStateEnumTypeWrapper): ...

DTLS_TRANSPORT_NEW: DtlsTransportState.ValueType  # 0
DTLS_TRANSPORT_CONNECTING: DtlsTransportState.ValueType  # 1
DTLS_TRANSPORT_CONNECTED: DtlsTransportState.ValueType  # 2
DTLS_TRANSPORT_CLOSED: DtlsTransportState.ValueType  # 3
DTLS_TRANSPORT_FAILED: DtlsTransportState.ValueType  # 4
global___DtlsTransportState = DtlsTransportState

class _IceTransportState:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _IceTransportStateEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_IceTransportState.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    ICE_TRANSPORT_NEW: _IceTransportState.ValueType  # 0
    ICE_TRANSPORT_CHECKING: _IceTransportState.ValueType  # 1
    ICE_TRANSPORT_CONNECTED: _IceTransportState.ValueType  # 2
    ICE_TRANSPORT_COMPLETED: _IceTransportState.ValueType  # 3
    ICE_TRANSPORT_DISCONNECTED: _IceTransportState.ValueType  # 4
    ICE_TRANSPORT_FAILED: _IceTransportState.ValueType  # 5
    ICE_TRANSPORT_CLOSED: _IceTransportState.ValueType  # 6

class IceTransportState(_IceTransportState, metaclass=_IceTransportStateEnumTypeWrapper): ...

ICE_TRANSPORT_NEW: IceTransportState.ValueType  # 0
ICE_TRANSPORT_CHECKING: IceTransportState.ValueType  # 1
ICE_TRANSPORT_CONNECTED: IceTransportState.ValueType  # 2
ICE_TRANSPORT_COMPLETED: IceTransportState.ValueType  # 3
ICE_TRANSPORT_DISCONNECTED: IceTransportState.ValueType  # 4
ICE_TRANSPORT_FAILED: IceTransportState.ValueType  # 5
ICE_TRANSPORT_CLOSED: IceTransportState.ValueType  # 6
global___IceTransportState = IceTransportState

class _DtlsRole:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _DtlsRoleEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_DtlsRole.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    DTLS_CLIENT: _DtlsRole.ValueType  # 0
    DTLS_SERVER: _DtlsRole.ValueType  # 1
    DTLS_UNKNOWN: _DtlsRole.ValueType  # 2

class DtlsRole(_DtlsRole, metaclass=_DtlsRoleEnumTypeWrapper): ...

DTLS_CLIENT: DtlsRole.ValueType  # 0
DTLS_SERVER: DtlsRole.ValueType  # 1
DTLS_UNKNOWN: DtlsRole.ValueType  # 2
global___DtlsRole = DtlsRole

class _IceCandidatePairState:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _IceCandidatePairStateEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_IceCandidatePairState.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    PAIR_FROZEN: _IceCandidatePairState.ValueType  # 0
    PAIR_WAITING: _IceCandidatePairState.ValueType  # 1
    PAIR_IN_PROGRESS: _IceCandidatePairState.ValueType  # 2
    PAIR_FAILED: _IceCandidatePairState.ValueType  # 3
    PAIR_SUCCEEDED: _IceCandidatePairState.ValueType  # 4

class IceCandidatePairState(_IceCandidatePairState, metaclass=_IceCandidatePairStateEnumTypeWrapper): ...

PAIR_FROZEN: IceCandidatePairState.ValueType  # 0
PAIR_WAITING: IceCandidatePairState.ValueType  # 1
PAIR_IN_PROGRESS: IceCandidatePairState.ValueType  # 2
PAIR_FAILED: IceCandidatePairState.ValueType  # 3
PAIR_SUCCEEDED: IceCandidatePairState.ValueType  # 4
global___IceCandidatePairState = IceCandidatePairState

class _IceCandidateType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _IceCandidateTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_IceCandidateType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    HOST: _IceCandidateType.ValueType  # 0
    SRFLX: _IceCandidateType.ValueType  # 1
    PRFLX: _IceCandidateType.ValueType  # 2
    RELAY: _IceCandidateType.ValueType  # 3

class IceCandidateType(_IceCandidateType, metaclass=_IceCandidateTypeEnumTypeWrapper): ...

HOST: IceCandidateType.ValueType  # 0
SRFLX: IceCandidateType.ValueType  # 1
PRFLX: IceCandidateType.ValueType  # 2
RELAY: IceCandidateType.ValueType  # 3
global___IceCandidateType = IceCandidateType

class _IceServerTransportProtocol:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _IceServerTransportProtocolEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_IceServerTransportProtocol.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    TRANSPORT_UDP: _IceServerTransportProtocol.ValueType  # 0
    TRANSPORT_TCP: _IceServerTransportProtocol.ValueType  # 1
    TRANSPORT_TLS: _IceServerTransportProtocol.ValueType  # 2

class IceServerTransportProtocol(_IceServerTransportProtocol, metaclass=_IceServerTransportProtocolEnumTypeWrapper): ...

TRANSPORT_UDP: IceServerTransportProtocol.ValueType  # 0
TRANSPORT_TCP: IceServerTransportProtocol.ValueType  # 1
TRANSPORT_TLS: IceServerTransportProtocol.ValueType  # 2
global___IceServerTransportProtocol = IceServerTransportProtocol

class _IceTcpCandidateType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _IceTcpCandidateTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_IceTcpCandidateType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    CANDIDATE_ACTIVE: _IceTcpCandidateType.ValueType  # 0
    CANDIDATE_PASSIVE: _IceTcpCandidateType.ValueType  # 1
    CANDIDATE_SO: _IceTcpCandidateType.ValueType  # 2

class IceTcpCandidateType(_IceTcpCandidateType, metaclass=_IceTcpCandidateTypeEnumTypeWrapper): ...

CANDIDATE_ACTIVE: IceTcpCandidateType.ValueType  # 0
CANDIDATE_PASSIVE: IceTcpCandidateType.ValueType  # 1
CANDIDATE_SO: IceTcpCandidateType.ValueType  # 2
global___IceTcpCandidateType = IceTcpCandidateType

@typing.final
class RtcStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Codec(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        CODEC_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def codec(self) -> global___CodecStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            codec: global___CodecStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["codec", b"codec", "rtc", b"rtc"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["codec", b"codec", "rtc", b"rtc"]) -> None: ...

    @typing.final
    class InboundRtp(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        STREAM_FIELD_NUMBER: builtins.int
        RECEIVED_FIELD_NUMBER: builtins.int
        INBOUND_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def stream(self) -> global___RtpStreamStats: ...
        @property
        def received(self) -> global___ReceivedRtpStreamStats: ...
        @property
        def inbound(self) -> global___InboundRtpStreamStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            stream: global___RtpStreamStats | None = ...,
            received: global___ReceivedRtpStreamStats | None = ...,
            inbound: global___InboundRtpStreamStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["inbound", b"inbound", "received", b"received", "rtc", b"rtc", "stream", b"stream"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["inbound", b"inbound", "received", b"received", "rtc", b"rtc", "stream", b"stream"]) -> None: ...

    @typing.final
    class OutboundRtp(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        STREAM_FIELD_NUMBER: builtins.int
        SENT_FIELD_NUMBER: builtins.int
        OUTBOUND_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def stream(self) -> global___RtpStreamStats: ...
        @property
        def sent(self) -> global___SentRtpStreamStats: ...
        @property
        def outbound(self) -> global___OutboundRtpStreamStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            stream: global___RtpStreamStats | None = ...,
            sent: global___SentRtpStreamStats | None = ...,
            outbound: global___OutboundRtpStreamStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["outbound", b"outbound", "rtc", b"rtc", "sent", b"sent", "stream", b"stream"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["outbound", b"outbound", "rtc", b"rtc", "sent", b"sent", "stream", b"stream"]) -> None: ...

    @typing.final
    class RemoteInboundRtp(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        STREAM_FIELD_NUMBER: builtins.int
        RECEIVED_FIELD_NUMBER: builtins.int
        REMOTE_INBOUND_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def stream(self) -> global___RtpStreamStats: ...
        @property
        def received(self) -> global___ReceivedRtpStreamStats: ...
        @property
        def remote_inbound(self) -> global___RemoteInboundRtpStreamStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            stream: global___RtpStreamStats | None = ...,
            received: global___ReceivedRtpStreamStats | None = ...,
            remote_inbound: global___RemoteInboundRtpStreamStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["received", b"received", "remote_inbound", b"remote_inbound", "rtc", b"rtc", "stream", b"stream"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["received", b"received", "remote_inbound", b"remote_inbound", "rtc", b"rtc", "stream", b"stream"]) -> None: ...

    @typing.final
    class RemoteOutboundRtp(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        STREAM_FIELD_NUMBER: builtins.int
        SENT_FIELD_NUMBER: builtins.int
        REMOTE_OUTBOUND_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def stream(self) -> global___RtpStreamStats: ...
        @property
        def sent(self) -> global___SentRtpStreamStats: ...
        @property
        def remote_outbound(self) -> global___RemoteOutboundRtpStreamStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            stream: global___RtpStreamStats | None = ...,
            sent: global___SentRtpStreamStats | None = ...,
            remote_outbound: global___RemoteOutboundRtpStreamStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["remote_outbound", b"remote_outbound", "rtc", b"rtc", "sent", b"sent", "stream", b"stream"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["remote_outbound", b"remote_outbound", "rtc", b"rtc", "sent", b"sent", "stream", b"stream"]) -> None: ...

    @typing.final
    class MediaSource(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        SOURCE_FIELD_NUMBER: builtins.int
        AUDIO_FIELD_NUMBER: builtins.int
        VIDEO_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def source(self) -> global___MediaSourceStats: ...
        @property
        def audio(self) -> global___AudioSourceStats: ...
        @property
        def video(self) -> global___VideoSourceStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            source: global___MediaSourceStats | None = ...,
            audio: global___AudioSourceStats | None = ...,
            video: global___VideoSourceStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["audio", b"audio", "rtc", b"rtc", "source", b"source", "video", b"video"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["audio", b"audio", "rtc", b"rtc", "source", b"source", "video", b"video"]) -> None: ...

    @typing.final
    class MediaPlayout(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        AUDIO_PLAYOUT_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def audio_playout(self) -> global___AudioPlayoutStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            audio_playout: global___AudioPlayoutStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["audio_playout", b"audio_playout", "rtc", b"rtc"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["audio_playout", b"audio_playout", "rtc", b"rtc"]) -> None: ...

    @typing.final
    class PeerConnection(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        PC_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def pc(self) -> global___PeerConnectionStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            pc: global___PeerConnectionStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["pc", b"pc", "rtc", b"rtc"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["pc", b"pc", "rtc", b"rtc"]) -> None: ...

    @typing.final
    class DataChannel(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        DC_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def dc(self) -> global___DataChannelStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            dc: global___DataChannelStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["dc", b"dc", "rtc", b"rtc"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["dc", b"dc", "rtc", b"rtc"]) -> None: ...

    @typing.final
    class Transport(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        TRANSPORT_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def transport(self) -> global___TransportStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            transport: global___TransportStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["rtc", b"rtc", "transport", b"transport"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["rtc", b"rtc", "transport", b"transport"]) -> None: ...

    @typing.final
    class CandidatePair(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        CANDIDATE_PAIR_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def candidate_pair(self) -> global___CandidatePairStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            candidate_pair: global___CandidatePairStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["candidate_pair", b"candidate_pair", "rtc", b"rtc"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["candidate_pair", b"candidate_pair", "rtc", b"rtc"]) -> None: ...

    @typing.final
    class LocalCandidate(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        CANDIDATE_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def candidate(self) -> global___IceCandidateStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            candidate: global___IceCandidateStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["candidate", b"candidate", "rtc", b"rtc"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["candidate", b"candidate", "rtc", b"rtc"]) -> None: ...

    @typing.final
    class RemoteCandidate(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        CANDIDATE_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def candidate(self) -> global___IceCandidateStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            candidate: global___IceCandidateStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["candidate", b"candidate", "rtc", b"rtc"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["candidate", b"candidate", "rtc", b"rtc"]) -> None: ...

    @typing.final
    class Certificate(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        CERTIFICATE_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def certificate(self) -> global___CertificateStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            certificate: global___CertificateStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["certificate", b"certificate", "rtc", b"rtc"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["certificate", b"certificate", "rtc", b"rtc"]) -> None: ...

    @typing.final
    class Stream(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RTC_FIELD_NUMBER: builtins.int
        STREAM_FIELD_NUMBER: builtins.int
        @property
        def rtc(self) -> global___RtcStatsData: ...
        @property
        def stream(self) -> global___StreamStats: ...
        def __init__(
            self,
            *,
            rtc: global___RtcStatsData | None = ...,
            stream: global___StreamStats | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["rtc", b"rtc", "stream", b"stream"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["rtc", b"rtc", "stream", b"stream"]) -> None: ...

    @typing.final
    class Track(google.protobuf.message.Message):
        """Deprecated"""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        def __init__(
            self,
        ) -> None: ...

    CODEC_FIELD_NUMBER: builtins.int
    INBOUND_RTP_FIELD_NUMBER: builtins.int
    OUTBOUND_RTP_FIELD_NUMBER: builtins.int
    REMOTE_INBOUND_RTP_FIELD_NUMBER: builtins.int
    REMOTE_OUTBOUND_RTP_FIELD_NUMBER: builtins.int
    MEDIA_SOURCE_FIELD_NUMBER: builtins.int
    MEDIA_PLAYOUT_FIELD_NUMBER: builtins.int
    PEER_CONNECTION_FIELD_NUMBER: builtins.int
    DATA_CHANNEL_FIELD_NUMBER: builtins.int
    TRANSPORT_FIELD_NUMBER: builtins.int
    CANDIDATE_PAIR_FIELD_NUMBER: builtins.int
    LOCAL_CANDIDATE_FIELD_NUMBER: builtins.int
    REMOTE_CANDIDATE_FIELD_NUMBER: builtins.int
    CERTIFICATE_FIELD_NUMBER: builtins.int
    STREAM_FIELD_NUMBER: builtins.int
    TRACK_FIELD_NUMBER: builtins.int
    @property
    def codec(self) -> global___RtcStats.Codec: ...
    @property
    def inbound_rtp(self) -> global___RtcStats.InboundRtp: ...
    @property
    def outbound_rtp(self) -> global___RtcStats.OutboundRtp: ...
    @property
    def remote_inbound_rtp(self) -> global___RtcStats.RemoteInboundRtp: ...
    @property
    def remote_outbound_rtp(self) -> global___RtcStats.RemoteOutboundRtp: ...
    @property
    def media_source(self) -> global___RtcStats.MediaSource: ...
    @property
    def media_playout(self) -> global___RtcStats.MediaPlayout: ...
    @property
    def peer_connection(self) -> global___RtcStats.PeerConnection: ...
    @property
    def data_channel(self) -> global___RtcStats.DataChannel: ...
    @property
    def transport(self) -> global___RtcStats.Transport: ...
    @property
    def candidate_pair(self) -> global___RtcStats.CandidatePair: ...
    @property
    def local_candidate(self) -> global___RtcStats.LocalCandidate: ...
    @property
    def remote_candidate(self) -> global___RtcStats.RemoteCandidate: ...
    @property
    def certificate(self) -> global___RtcStats.Certificate: ...
    @property
    def stream(self) -> global___RtcStats.Stream: ...
    @property
    def track(self) -> global___RtcStats.Track: ...
    def __init__(
        self,
        *,
        codec: global___RtcStats.Codec | None = ...,
        inbound_rtp: global___RtcStats.InboundRtp | None = ...,
        outbound_rtp: global___RtcStats.OutboundRtp | None = ...,
        remote_inbound_rtp: global___RtcStats.RemoteInboundRtp | None = ...,
        remote_outbound_rtp: global___RtcStats.RemoteOutboundRtp | None = ...,
        media_source: global___RtcStats.MediaSource | None = ...,
        media_playout: global___RtcStats.MediaPlayout | None = ...,
        peer_connection: global___RtcStats.PeerConnection | None = ...,
        data_channel: global___RtcStats.DataChannel | None = ...,
        transport: global___RtcStats.Transport | None = ...,
        candidate_pair: global___RtcStats.CandidatePair | None = ...,
        local_candidate: global___RtcStats.LocalCandidate | None = ...,
        remote_candidate: global___RtcStats.RemoteCandidate | None = ...,
        certificate: global___RtcStats.Certificate | None = ...,
        stream: global___RtcStats.Stream | None = ...,
        track: global___RtcStats.Track | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["candidate_pair", b"candidate_pair", "certificate", b"certificate", "codec", b"codec", "data_channel", b"data_channel", "inbound_rtp", b"inbound_rtp", "local_candidate", b"local_candidate", "media_playout", b"media_playout", "media_source", b"media_source", "outbound_rtp", b"outbound_rtp", "peer_connection", b"peer_connection", "remote_candidate", b"remote_candidate", "remote_inbound_rtp", b"remote_inbound_rtp", "remote_outbound_rtp", b"remote_outbound_rtp", "stats", b"stats", "stream", b"stream", "track", b"track", "transport", b"transport"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["candidate_pair", b"candidate_pair", "certificate", b"certificate", "codec", b"codec", "data_channel", b"data_channel", "inbound_rtp", b"inbound_rtp", "local_candidate", b"local_candidate", "media_playout", b"media_playout", "media_source", b"media_source", "outbound_rtp", b"outbound_rtp", "peer_connection", b"peer_connection", "remote_candidate", b"remote_candidate", "remote_inbound_rtp", b"remote_inbound_rtp", "remote_outbound_rtp", b"remote_outbound_rtp", "stats", b"stats", "stream", b"stream", "track", b"track", "transport", b"transport"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["stats", b"stats"]) -> typing.Literal["codec", "inbound_rtp", "outbound_rtp", "remote_inbound_rtp", "remote_outbound_rtp", "media_source", "media_playout", "peer_connection", "data_channel", "transport", "candidate_pair", "local_candidate", "remote_candidate", "certificate", "stream", "track"] | None: ...

global___RtcStats = RtcStats

@typing.final
class RtcStatsData(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    TIMESTAMP_FIELD_NUMBER: builtins.int
    id: builtins.str
    timestamp: builtins.int
    def __init__(
        self,
        *,
        id: builtins.str | None = ...,
        timestamp: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["id", b"id", "timestamp", b"timestamp"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["id", b"id", "timestamp", b"timestamp"]) -> None: ...

global___RtcStatsData = RtcStatsData

@typing.final
class CodecStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PAYLOAD_TYPE_FIELD_NUMBER: builtins.int
    TRANSPORT_ID_FIELD_NUMBER: builtins.int
    MIME_TYPE_FIELD_NUMBER: builtins.int
    CLOCK_RATE_FIELD_NUMBER: builtins.int
    CHANNELS_FIELD_NUMBER: builtins.int
    SDP_FMTP_LINE_FIELD_NUMBER: builtins.int
    payload_type: builtins.int
    transport_id: builtins.str
    mime_type: builtins.str
    clock_rate: builtins.int
    channels: builtins.int
    sdp_fmtp_line: builtins.str
    def __init__(
        self,
        *,
        payload_type: builtins.int | None = ...,
        transport_id: builtins.str | None = ...,
        mime_type: builtins.str | None = ...,
        clock_rate: builtins.int | None = ...,
        channels: builtins.int | None = ...,
        sdp_fmtp_line: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["channels", b"channels", "clock_rate", b"clock_rate", "mime_type", b"mime_type", "payload_type", b"payload_type", "sdp_fmtp_line", b"sdp_fmtp_line", "transport_id", b"transport_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["channels", b"channels", "clock_rate", b"clock_rate", "mime_type", b"mime_type", "payload_type", b"payload_type", "sdp_fmtp_line", b"sdp_fmtp_line", "transport_id", b"transport_id"]) -> None: ...

global___CodecStats = CodecStats

@typing.final
class RtpStreamStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SSRC_FIELD_NUMBER: builtins.int
    KIND_FIELD_NUMBER: builtins.int
    TRANSPORT_ID_FIELD_NUMBER: builtins.int
    CODEC_ID_FIELD_NUMBER: builtins.int
    ssrc: builtins.int
    kind: builtins.str
    transport_id: builtins.str
    codec_id: builtins.str
    def __init__(
        self,
        *,
        ssrc: builtins.int | None = ...,
        kind: builtins.str | None = ...,
        transport_id: builtins.str | None = ...,
        codec_id: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["codec_id", b"codec_id", "kind", b"kind", "ssrc", b"ssrc", "transport_id", b"transport_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["codec_id", b"codec_id", "kind", b"kind", "ssrc", b"ssrc", "transport_id", b"transport_id"]) -> None: ...

global___RtpStreamStats = RtpStreamStats

@typing.final
class ReceivedRtpStreamStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PACKETS_RECEIVED_FIELD_NUMBER: builtins.int
    PACKETS_LOST_FIELD_NUMBER: builtins.int
    JITTER_FIELD_NUMBER: builtins.int
    packets_received: builtins.int
    packets_lost: builtins.int
    jitter: builtins.float
    def __init__(
        self,
        *,
        packets_received: builtins.int | None = ...,
        packets_lost: builtins.int | None = ...,
        jitter: builtins.float | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["jitter", b"jitter", "packets_lost", b"packets_lost", "packets_received", b"packets_received"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["jitter", b"jitter", "packets_lost", b"packets_lost", "packets_received", b"packets_received"]) -> None: ...

global___ReceivedRtpStreamStats = ReceivedRtpStreamStats

@typing.final
class InboundRtpStreamStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TRACK_IDENTIFIER_FIELD_NUMBER: builtins.int
    MID_FIELD_NUMBER: builtins.int
    REMOTE_ID_FIELD_NUMBER: builtins.int
    FRAMES_DECODED_FIELD_NUMBER: builtins.int
    KEY_FRAMES_DECODED_FIELD_NUMBER: builtins.int
    FRAMES_RENDERED_FIELD_NUMBER: builtins.int
    FRAMES_DROPPED_FIELD_NUMBER: builtins.int
    FRAME_WIDTH_FIELD_NUMBER: builtins.int
    FRAME_HEIGHT_FIELD_NUMBER: builtins.int
    FRAMES_PER_SECOND_FIELD_NUMBER: builtins.int
    QP_SUM_FIELD_NUMBER: builtins.int
    TOTAL_DECODE_TIME_FIELD_NUMBER: builtins.int
    TOTAL_INTER_FRAME_DELAY_FIELD_NUMBER: builtins.int
    TOTAL_SQUARED_INTER_FRAME_DELAY_FIELD_NUMBER: builtins.int
    PAUSE_COUNT_FIELD_NUMBER: builtins.int
    TOTAL_PAUSE_DURATION_FIELD_NUMBER: builtins.int
    FREEZE_COUNT_FIELD_NUMBER: builtins.int
    TOTAL_FREEZE_DURATION_FIELD_NUMBER: builtins.int
    LAST_PACKET_RECEIVED_TIMESTAMP_FIELD_NUMBER: builtins.int
    HEADER_BYTES_RECEIVED_FIELD_NUMBER: builtins.int
    PACKETS_DISCARDED_FIELD_NUMBER: builtins.int
    FEC_BYTES_RECEIVED_FIELD_NUMBER: builtins.int
    FEC_PACKETS_RECEIVED_FIELD_NUMBER: builtins.int
    FEC_PACKETS_DISCARDED_FIELD_NUMBER: builtins.int
    BYTES_RECEIVED_FIELD_NUMBER: builtins.int
    NACK_COUNT_FIELD_NUMBER: builtins.int
    FIR_COUNT_FIELD_NUMBER: builtins.int
    PLI_COUNT_FIELD_NUMBER: builtins.int
    TOTAL_PROCESSING_DELAY_FIELD_NUMBER: builtins.int
    ESTIMATED_PLAYOUT_TIMESTAMP_FIELD_NUMBER: builtins.int
    JITTER_BUFFER_DELAY_FIELD_NUMBER: builtins.int
    JITTER_BUFFER_TARGET_DELAY_FIELD_NUMBER: builtins.int
    JITTER_BUFFER_EMITTED_COUNT_FIELD_NUMBER: builtins.int
    JITTER_BUFFER_MINIMUM_DELAY_FIELD_NUMBER: builtins.int
    TOTAL_SAMPLES_RECEIVED_FIELD_NUMBER: builtins.int
    CONCEALED_SAMPLES_FIELD_NUMBER: builtins.int
    SILENT_CONCEALED_SAMPLES_FIELD_NUMBER: builtins.int
    CONCEALMENT_EVENTS_FIELD_NUMBER: builtins.int
    INSERTED_SAMPLES_FOR_DECELERATION_FIELD_NUMBER: builtins.int
    REMOVED_SAMPLES_FOR_ACCELERATION_FIELD_NUMBER: builtins.int
    AUDIO_LEVEL_FIELD_NUMBER: builtins.int
    TOTAL_AUDIO_ENERGY_FIELD_NUMBER: builtins.int
    TOTAL_SAMPLES_DURATION_FIELD_NUMBER: builtins.int
    FRAMES_RECEIVED_FIELD_NUMBER: builtins.int
    DECODER_IMPLEMENTATION_FIELD_NUMBER: builtins.int
    PLAYOUT_ID_FIELD_NUMBER: builtins.int
    POWER_EFFICIENT_DECODER_FIELD_NUMBER: builtins.int
    FRAMES_ASSEMBLED_FROM_MULTIPLE_PACKETS_FIELD_NUMBER: builtins.int
    TOTAL_ASSEMBLY_TIME_FIELD_NUMBER: builtins.int
    RETRANSMITTED_PACKETS_RECEIVED_FIELD_NUMBER: builtins.int
    RETRANSMITTED_BYTES_RECEIVED_FIELD_NUMBER: builtins.int
    RTX_SSRC_FIELD_NUMBER: builtins.int
    FEC_SSRC_FIELD_NUMBER: builtins.int
    track_identifier: builtins.str
    mid: builtins.str
    remote_id: builtins.str
    frames_decoded: builtins.int
    key_frames_decoded: builtins.int
    frames_rendered: builtins.int
    frames_dropped: builtins.int
    frame_width: builtins.int
    frame_height: builtins.int
    frames_per_second: builtins.float
    qp_sum: builtins.int
    total_decode_time: builtins.float
    total_inter_frame_delay: builtins.float
    total_squared_inter_frame_delay: builtins.float
    pause_count: builtins.int
    total_pause_duration: builtins.float
    freeze_count: builtins.int
    total_freeze_duration: builtins.float
    last_packet_received_timestamp: builtins.float
    header_bytes_received: builtins.int
    packets_discarded: builtins.int
    fec_bytes_received: builtins.int
    fec_packets_received: builtins.int
    fec_packets_discarded: builtins.int
    bytes_received: builtins.int
    nack_count: builtins.int
    fir_count: builtins.int
    pli_count: builtins.int
    total_processing_delay: builtins.float
    estimated_playout_timestamp: builtins.float
    jitter_buffer_delay: builtins.float
    jitter_buffer_target_delay: builtins.float
    jitter_buffer_emitted_count: builtins.int
    jitter_buffer_minimum_delay: builtins.float
    total_samples_received: builtins.int
    concealed_samples: builtins.int
    silent_concealed_samples: builtins.int
    concealment_events: builtins.int
    inserted_samples_for_deceleration: builtins.int
    removed_samples_for_acceleration: builtins.int
    audio_level: builtins.float
    total_audio_energy: builtins.float
    total_samples_duration: builtins.float
    frames_received: builtins.int
    decoder_implementation: builtins.str
    playout_id: builtins.str
    power_efficient_decoder: builtins.bool
    frames_assembled_from_multiple_packets: builtins.int
    total_assembly_time: builtins.float
    retransmitted_packets_received: builtins.int
    retransmitted_bytes_received: builtins.int
    rtx_ssrc: builtins.int
    fec_ssrc: builtins.int
    def __init__(
        self,
        *,
        track_identifier: builtins.str | None = ...,
        mid: builtins.str | None = ...,
        remote_id: builtins.str | None = ...,
        frames_decoded: builtins.int | None = ...,
        key_frames_decoded: builtins.int | None = ...,
        frames_rendered: builtins.int | None = ...,
        frames_dropped: builtins.int | None = ...,
        frame_width: builtins.int | None = ...,
        frame_height: builtins.int | None = ...,
        frames_per_second: builtins.float | None = ...,
        qp_sum: builtins.int | None = ...,
        total_decode_time: builtins.float | None = ...,
        total_inter_frame_delay: builtins.float | None = ...,
        total_squared_inter_frame_delay: builtins.float | None = ...,
        pause_count: builtins.int | None = ...,
        total_pause_duration: builtins.float | None = ...,
        freeze_count: builtins.int | None = ...,
        total_freeze_duration: builtins.float | None = ...,
        last_packet_received_timestamp: builtins.float | None = ...,
        header_bytes_received: builtins.int | None = ...,
        packets_discarded: builtins.int | None = ...,
        fec_bytes_received: builtins.int | None = ...,
        fec_packets_received: builtins.int | None = ...,
        fec_packets_discarded: builtins.int | None = ...,
        bytes_received: builtins.int | None = ...,
        nack_count: builtins.int | None = ...,
        fir_count: builtins.int | None = ...,
        pli_count: builtins.int | None = ...,
        total_processing_delay: builtins.float | None = ...,
        estimated_playout_timestamp: builtins.float | None = ...,
        jitter_buffer_delay: builtins.float | None = ...,
        jitter_buffer_target_delay: builtins.float | None = ...,
        jitter_buffer_emitted_count: builtins.int | None = ...,
        jitter_buffer_minimum_delay: builtins.float | None = ...,
        total_samples_received: builtins.int | None = ...,
        concealed_samples: builtins.int | None = ...,
        silent_concealed_samples: builtins.int | None = ...,
        concealment_events: builtins.int | None = ...,
        inserted_samples_for_deceleration: builtins.int | None = ...,
        removed_samples_for_acceleration: builtins.int | None = ...,
        audio_level: builtins.float | None = ...,
        total_audio_energy: builtins.float | None = ...,
        total_samples_duration: builtins.float | None = ...,
        frames_received: builtins.int | None = ...,
        decoder_implementation: builtins.str | None = ...,
        playout_id: builtins.str | None = ...,
        power_efficient_decoder: builtins.bool | None = ...,
        frames_assembled_from_multiple_packets: builtins.int | None = ...,
        total_assembly_time: builtins.float | None = ...,
        retransmitted_packets_received: builtins.int | None = ...,
        retransmitted_bytes_received: builtins.int | None = ...,
        rtx_ssrc: builtins.int | None = ...,
        fec_ssrc: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["audio_level", b"audio_level", "bytes_received", b"bytes_received", "concealed_samples", b"concealed_samples", "concealment_events", b"concealment_events", "decoder_implementation", b"decoder_implementation", "estimated_playout_timestamp", b"estimated_playout_timestamp", "fec_bytes_received", b"fec_bytes_received", "fec_packets_discarded", b"fec_packets_discarded", "fec_packets_received", b"fec_packets_received", "fec_ssrc", b"fec_ssrc", "fir_count", b"fir_count", "frame_height", b"frame_height", "frame_width", b"frame_width", "frames_assembled_from_multiple_packets", b"frames_assembled_from_multiple_packets", "frames_decoded", b"frames_decoded", "frames_dropped", b"frames_dropped", "frames_per_second", b"frames_per_second", "frames_received", b"frames_received", "frames_rendered", b"frames_rendered", "freeze_count", b"freeze_count", "header_bytes_received", b"header_bytes_received", "inserted_samples_for_deceleration", b"inserted_samples_for_deceleration", "jitter_buffer_delay", b"jitter_buffer_delay", "jitter_buffer_emitted_count", b"jitter_buffer_emitted_count", "jitter_buffer_minimum_delay", b"jitter_buffer_minimum_delay", "jitter_buffer_target_delay", b"jitter_buffer_target_delay", "key_frames_decoded", b"key_frames_decoded", "last_packet_received_timestamp", b"last_packet_received_timestamp", "mid", b"mid", "nack_count", b"nack_count", "packets_discarded", b"packets_discarded", "pause_count", b"pause_count", "playout_id", b"playout_id", "pli_count", b"pli_count", "power_efficient_decoder", b"power_efficient_decoder", "qp_sum", b"qp_sum", "remote_id", b"remote_id", "removed_samples_for_acceleration", b"removed_samples_for_acceleration", "retransmitted_bytes_received", b"retransmitted_bytes_received", "retransmitted_packets_received", b"retransmitted_packets_received", "rtx_ssrc", b"rtx_ssrc", "silent_concealed_samples", b"silent_concealed_samples", "total_assembly_time", b"total_assembly_time", "total_audio_energy", b"total_audio_energy", "total_decode_time", b"total_decode_time", "total_freeze_duration", b"total_freeze_duration", "total_inter_frame_delay", b"total_inter_frame_delay", "total_pause_duration", b"total_pause_duration", "total_processing_delay", b"total_processing_delay", "total_samples_duration", b"total_samples_duration", "total_samples_received", b"total_samples_received", "total_squared_inter_frame_delay", b"total_squared_inter_frame_delay", "track_identifier", b"track_identifier"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["audio_level", b"audio_level", "bytes_received", b"bytes_received", "concealed_samples", b"concealed_samples", "concealment_events", b"concealment_events", "decoder_implementation", b"decoder_implementation", "estimated_playout_timestamp", b"estimated_playout_timestamp", "fec_bytes_received", b"fec_bytes_received", "fec_packets_discarded", b"fec_packets_discarded", "fec_packets_received", b"fec_packets_received", "fec_ssrc", b"fec_ssrc", "fir_count", b"fir_count", "frame_height", b"frame_height", "frame_width", b"frame_width", "frames_assembled_from_multiple_packets", b"frames_assembled_from_multiple_packets", "frames_decoded", b"frames_decoded", "frames_dropped", b"frames_dropped", "frames_per_second", b"frames_per_second", "frames_received", b"frames_received", "frames_rendered", b"frames_rendered", "freeze_count", b"freeze_count", "header_bytes_received", b"header_bytes_received", "inserted_samples_for_deceleration", b"inserted_samples_for_deceleration", "jitter_buffer_delay", b"jitter_buffer_delay", "jitter_buffer_emitted_count", b"jitter_buffer_emitted_count", "jitter_buffer_minimum_delay", b"jitter_buffer_minimum_delay", "jitter_buffer_target_delay", b"jitter_buffer_target_delay", "key_frames_decoded", b"key_frames_decoded", "last_packet_received_timestamp", b"last_packet_received_timestamp", "mid", b"mid", "nack_count", b"nack_count", "packets_discarded", b"packets_discarded", "pause_count", b"pause_count", "playout_id", b"playout_id", "pli_count", b"pli_count", "power_efficient_decoder", b"power_efficient_decoder", "qp_sum", b"qp_sum", "remote_id", b"remote_id", "removed_samples_for_acceleration", b"removed_samples_for_acceleration", "retransmitted_bytes_received", b"retransmitted_bytes_received", "retransmitted_packets_received", b"retransmitted_packets_received", "rtx_ssrc", b"rtx_ssrc", "silent_concealed_samples", b"silent_concealed_samples", "total_assembly_time", b"total_assembly_time", "total_audio_energy", b"total_audio_energy", "total_decode_time", b"total_decode_time", "total_freeze_duration", b"total_freeze_duration", "total_inter_frame_delay", b"total_inter_frame_delay", "total_pause_duration", b"total_pause_duration", "total_processing_delay", b"total_processing_delay", "total_samples_duration", b"total_samples_duration", "total_samples_received", b"total_samples_received", "total_squared_inter_frame_delay", b"total_squared_inter_frame_delay", "track_identifier", b"track_identifier"]) -> None: ...

global___InboundRtpStreamStats = InboundRtpStreamStats

@typing.final
class SentRtpStreamStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PACKETS_SENT_FIELD_NUMBER: builtins.int
    BYTES_SENT_FIELD_NUMBER: builtins.int
    packets_sent: builtins.int
    bytes_sent: builtins.int
    def __init__(
        self,
        *,
        packets_sent: builtins.int | None = ...,
        bytes_sent: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bytes_sent", b"bytes_sent", "packets_sent", b"packets_sent"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bytes_sent", b"bytes_sent", "packets_sent", b"packets_sent"]) -> None: ...

global___SentRtpStreamStats = SentRtpStreamStats

@typing.final
class OutboundRtpStreamStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class QualityLimitationDurationsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.float
        def __init__(
            self,
            *,
            key: builtins.str | None = ...,
            value: builtins.float | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    MID_FIELD_NUMBER: builtins.int
    MEDIA_SOURCE_ID_FIELD_NUMBER: builtins.int
    REMOTE_ID_FIELD_NUMBER: builtins.int
    RID_FIELD_NUMBER: builtins.int
    HEADER_BYTES_SENT_FIELD_NUMBER: builtins.int
    RETRANSMITTED_PACKETS_SENT_FIELD_NUMBER: builtins.int
    RETRANSMITTED_BYTES_SENT_FIELD_NUMBER: builtins.int
    RTX_SSRC_FIELD_NUMBER: builtins.int
    TARGET_BITRATE_FIELD_NUMBER: builtins.int
    TOTAL_ENCODED_BYTES_TARGET_FIELD_NUMBER: builtins.int
    FRAME_WIDTH_FIELD_NUMBER: builtins.int
    FRAME_HEIGHT_FIELD_NUMBER: builtins.int
    FRAMES_PER_SECOND_FIELD_NUMBER: builtins.int
    FRAMES_SENT_FIELD_NUMBER: builtins.int
    HUGE_FRAMES_SENT_FIELD_NUMBER: builtins.int
    FRAMES_ENCODED_FIELD_NUMBER: builtins.int
    KEY_FRAMES_ENCODED_FIELD_NUMBER: builtins.int
    QP_SUM_FIELD_NUMBER: builtins.int
    TOTAL_ENCODE_TIME_FIELD_NUMBER: builtins.int
    TOTAL_PACKET_SEND_DELAY_FIELD_NUMBER: builtins.int
    QUALITY_LIMITATION_REASON_FIELD_NUMBER: builtins.int
    QUALITY_LIMITATION_DURATIONS_FIELD_NUMBER: builtins.int
    QUALITY_LIMITATION_RESOLUTION_CHANGES_FIELD_NUMBER: builtins.int
    NACK_COUNT_FIELD_NUMBER: builtins.int
    FIR_COUNT_FIELD_NUMBER: builtins.int
    PLI_COUNT_FIELD_NUMBER: builtins.int
    ENCODER_IMPLEMENTATION_FIELD_NUMBER: builtins.int
    POWER_EFFICIENT_ENCODER_FIELD_NUMBER: builtins.int
    ACTIVE_FIELD_NUMBER: builtins.int
    SCALABILITY_MODE_FIELD_NUMBER: builtins.int
    mid: builtins.str
    media_source_id: builtins.str
    remote_id: builtins.str
    rid: builtins.str
    header_bytes_sent: builtins.int
    retransmitted_packets_sent: builtins.int
    retransmitted_bytes_sent: builtins.int
    rtx_ssrc: builtins.int
    target_bitrate: builtins.float
    total_encoded_bytes_target: builtins.int
    frame_width: builtins.int
    frame_height: builtins.int
    frames_per_second: builtins.float
    frames_sent: builtins.int
    huge_frames_sent: builtins.int
    frames_encoded: builtins.int
    key_frames_encoded: builtins.int
    qp_sum: builtins.int
    total_encode_time: builtins.float
    total_packet_send_delay: builtins.float
    quality_limitation_reason: global___QualityLimitationReason.ValueType
    quality_limitation_resolution_changes: builtins.int
    nack_count: builtins.int
    fir_count: builtins.int
    pli_count: builtins.int
    encoder_implementation: builtins.str
    power_efficient_encoder: builtins.bool
    active: builtins.bool
    scalability_mode: builtins.str
    @property
    def quality_limitation_durations(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.float]: ...
    def __init__(
        self,
        *,
        mid: builtins.str | None = ...,
        media_source_id: builtins.str | None = ...,
        remote_id: builtins.str | None = ...,
        rid: builtins.str | None = ...,
        header_bytes_sent: builtins.int | None = ...,
        retransmitted_packets_sent: builtins.int | None = ...,
        retransmitted_bytes_sent: builtins.int | None = ...,
        rtx_ssrc: builtins.int | None = ...,
        target_bitrate: builtins.float | None = ...,
        total_encoded_bytes_target: builtins.int | None = ...,
        frame_width: builtins.int | None = ...,
        frame_height: builtins.int | None = ...,
        frames_per_second: builtins.float | None = ...,
        frames_sent: builtins.int | None = ...,
        huge_frames_sent: builtins.int | None = ...,
        frames_encoded: builtins.int | None = ...,
        key_frames_encoded: builtins.int | None = ...,
        qp_sum: builtins.int | None = ...,
        total_encode_time: builtins.float | None = ...,
        total_packet_send_delay: builtins.float | None = ...,
        quality_limitation_reason: global___QualityLimitationReason.ValueType | None = ...,
        quality_limitation_durations: collections.abc.Mapping[builtins.str, builtins.float] | None = ...,
        quality_limitation_resolution_changes: builtins.int | None = ...,
        nack_count: builtins.int | None = ...,
        fir_count: builtins.int | None = ...,
        pli_count: builtins.int | None = ...,
        encoder_implementation: builtins.str | None = ...,
        power_efficient_encoder: builtins.bool | None = ...,
        active: builtins.bool | None = ...,
        scalability_mode: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["active", b"active", "encoder_implementation", b"encoder_implementation", "fir_count", b"fir_count", "frame_height", b"frame_height", "frame_width", b"frame_width", "frames_encoded", b"frames_encoded", "frames_per_second", b"frames_per_second", "frames_sent", b"frames_sent", "header_bytes_sent", b"header_bytes_sent", "huge_frames_sent", b"huge_frames_sent", "key_frames_encoded", b"key_frames_encoded", "media_source_id", b"media_source_id", "mid", b"mid", "nack_count", b"nack_count", "pli_count", b"pli_count", "power_efficient_encoder", b"power_efficient_encoder", "qp_sum", b"qp_sum", "quality_limitation_reason", b"quality_limitation_reason", "quality_limitation_resolution_changes", b"quality_limitation_resolution_changes", "remote_id", b"remote_id", "retransmitted_bytes_sent", b"retransmitted_bytes_sent", "retransmitted_packets_sent", b"retransmitted_packets_sent", "rid", b"rid", "rtx_ssrc", b"rtx_ssrc", "scalability_mode", b"scalability_mode", "target_bitrate", b"target_bitrate", "total_encode_time", b"total_encode_time", "total_encoded_bytes_target", b"total_encoded_bytes_target", "total_packet_send_delay", b"total_packet_send_delay"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["active", b"active", "encoder_implementation", b"encoder_implementation", "fir_count", b"fir_count", "frame_height", b"frame_height", "frame_width", b"frame_width", "frames_encoded", b"frames_encoded", "frames_per_second", b"frames_per_second", "frames_sent", b"frames_sent", "header_bytes_sent", b"header_bytes_sent", "huge_frames_sent", b"huge_frames_sent", "key_frames_encoded", b"key_frames_encoded", "media_source_id", b"media_source_id", "mid", b"mid", "nack_count", b"nack_count", "pli_count", b"pli_count", "power_efficient_encoder", b"power_efficient_encoder", "qp_sum", b"qp_sum", "quality_limitation_durations", b"quality_limitation_durations", "quality_limitation_reason", b"quality_limitation_reason", "quality_limitation_resolution_changes", b"quality_limitation_resolution_changes", "remote_id", b"remote_id", "retransmitted_bytes_sent", b"retransmitted_bytes_sent", "retransmitted_packets_sent", b"retransmitted_packets_sent", "rid", b"rid", "rtx_ssrc", b"rtx_ssrc", "scalability_mode", b"scalability_mode", "target_bitrate", b"target_bitrate", "total_encode_time", b"total_encode_time", "total_encoded_bytes_target", b"total_encoded_bytes_target", "total_packet_send_delay", b"total_packet_send_delay"]) -> None: ...

global___OutboundRtpStreamStats = OutboundRtpStreamStats

@typing.final
class RemoteInboundRtpStreamStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LOCAL_ID_FIELD_NUMBER: builtins.int
    ROUND_TRIP_TIME_FIELD_NUMBER: builtins.int
    TOTAL_ROUND_TRIP_TIME_FIELD_NUMBER: builtins.int
    FRACTION_LOST_FIELD_NUMBER: builtins.int
    ROUND_TRIP_TIME_MEASUREMENTS_FIELD_NUMBER: builtins.int
    local_id: builtins.str
    round_trip_time: builtins.float
    total_round_trip_time: builtins.float
    fraction_lost: builtins.float
    round_trip_time_measurements: builtins.int
    def __init__(
        self,
        *,
        local_id: builtins.str | None = ...,
        round_trip_time: builtins.float | None = ...,
        total_round_trip_time: builtins.float | None = ...,
        fraction_lost: builtins.float | None = ...,
        round_trip_time_measurements: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["fraction_lost", b"fraction_lost", "local_id", b"local_id", "round_trip_time", b"round_trip_time", "round_trip_time_measurements", b"round_trip_time_measurements", "total_round_trip_time", b"total_round_trip_time"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["fraction_lost", b"fraction_lost", "local_id", b"local_id", "round_trip_time", b"round_trip_time", "round_trip_time_measurements", b"round_trip_time_measurements", "total_round_trip_time", b"total_round_trip_time"]) -> None: ...

global___RemoteInboundRtpStreamStats = RemoteInboundRtpStreamStats

@typing.final
class RemoteOutboundRtpStreamStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LOCAL_ID_FIELD_NUMBER: builtins.int
    REMOTE_TIMESTAMP_FIELD_NUMBER: builtins.int
    REPORTS_SENT_FIELD_NUMBER: builtins.int
    ROUND_TRIP_TIME_FIELD_NUMBER: builtins.int
    TOTAL_ROUND_TRIP_TIME_FIELD_NUMBER: builtins.int
    ROUND_TRIP_TIME_MEASUREMENTS_FIELD_NUMBER: builtins.int
    local_id: builtins.str
    remote_timestamp: builtins.float
    reports_sent: builtins.int
    round_trip_time: builtins.float
    total_round_trip_time: builtins.float
    round_trip_time_measurements: builtins.int
    def __init__(
        self,
        *,
        local_id: builtins.str | None = ...,
        remote_timestamp: builtins.float | None = ...,
        reports_sent: builtins.int | None = ...,
        round_trip_time: builtins.float | None = ...,
        total_round_trip_time: builtins.float | None = ...,
        round_trip_time_measurements: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["local_id", b"local_id", "remote_timestamp", b"remote_timestamp", "reports_sent", b"reports_sent", "round_trip_time", b"round_trip_time", "round_trip_time_measurements", b"round_trip_time_measurements", "total_round_trip_time", b"total_round_trip_time"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["local_id", b"local_id", "remote_timestamp", b"remote_timestamp", "reports_sent", b"reports_sent", "round_trip_time", b"round_trip_time", "round_trip_time_measurements", b"round_trip_time_measurements", "total_round_trip_time", b"total_round_trip_time"]) -> None: ...

global___RemoteOutboundRtpStreamStats = RemoteOutboundRtpStreamStats

@typing.final
class MediaSourceStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TRACK_IDENTIFIER_FIELD_NUMBER: builtins.int
    KIND_FIELD_NUMBER: builtins.int
    track_identifier: builtins.str
    kind: builtins.str
    def __init__(
        self,
        *,
        track_identifier: builtins.str | None = ...,
        kind: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["kind", b"kind", "track_identifier", b"track_identifier"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["kind", b"kind", "track_identifier", b"track_identifier"]) -> None: ...

global___MediaSourceStats = MediaSourceStats

@typing.final
class AudioSourceStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AUDIO_LEVEL_FIELD_NUMBER: builtins.int
    TOTAL_AUDIO_ENERGY_FIELD_NUMBER: builtins.int
    TOTAL_SAMPLES_DURATION_FIELD_NUMBER: builtins.int
    ECHO_RETURN_LOSS_FIELD_NUMBER: builtins.int
    ECHO_RETURN_LOSS_ENHANCEMENT_FIELD_NUMBER: builtins.int
    DROPPED_SAMPLES_DURATION_FIELD_NUMBER: builtins.int
    DROPPED_SAMPLES_EVENTS_FIELD_NUMBER: builtins.int
    TOTAL_CAPTURE_DELAY_FIELD_NUMBER: builtins.int
    TOTAL_SAMPLES_CAPTURED_FIELD_NUMBER: builtins.int
    audio_level: builtins.float
    total_audio_energy: builtins.float
    total_samples_duration: builtins.float
    echo_return_loss: builtins.float
    echo_return_loss_enhancement: builtins.float
    dropped_samples_duration: builtins.float
    dropped_samples_events: builtins.int
    total_capture_delay: builtins.float
    total_samples_captured: builtins.int
    def __init__(
        self,
        *,
        audio_level: builtins.float | None = ...,
        total_audio_energy: builtins.float | None = ...,
        total_samples_duration: builtins.float | None = ...,
        echo_return_loss: builtins.float | None = ...,
        echo_return_loss_enhancement: builtins.float | None = ...,
        dropped_samples_duration: builtins.float | None = ...,
        dropped_samples_events: builtins.int | None = ...,
        total_capture_delay: builtins.float | None = ...,
        total_samples_captured: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["audio_level", b"audio_level", "dropped_samples_duration", b"dropped_samples_duration", "dropped_samples_events", b"dropped_samples_events", "echo_return_loss", b"echo_return_loss", "echo_return_loss_enhancement", b"echo_return_loss_enhancement", "total_audio_energy", b"total_audio_energy", "total_capture_delay", b"total_capture_delay", "total_samples_captured", b"total_samples_captured", "total_samples_duration", b"total_samples_duration"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["audio_level", b"audio_level", "dropped_samples_duration", b"dropped_samples_duration", "dropped_samples_events", b"dropped_samples_events", "echo_return_loss", b"echo_return_loss", "echo_return_loss_enhancement", b"echo_return_loss_enhancement", "total_audio_energy", b"total_audio_energy", "total_capture_delay", b"total_capture_delay", "total_samples_captured", b"total_samples_captured", "total_samples_duration", b"total_samples_duration"]) -> None: ...

global___AudioSourceStats = AudioSourceStats

@typing.final
class VideoSourceStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    WIDTH_FIELD_NUMBER: builtins.int
    HEIGHT_FIELD_NUMBER: builtins.int
    FRAMES_FIELD_NUMBER: builtins.int
    FRAMES_PER_SECOND_FIELD_NUMBER: builtins.int
    width: builtins.int
    height: builtins.int
    frames: builtins.int
    frames_per_second: builtins.float
    def __init__(
        self,
        *,
        width: builtins.int | None = ...,
        height: builtins.int | None = ...,
        frames: builtins.int | None = ...,
        frames_per_second: builtins.float | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["frames", b"frames", "frames_per_second", b"frames_per_second", "height", b"height", "width", b"width"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["frames", b"frames", "frames_per_second", b"frames_per_second", "height", b"height", "width", b"width"]) -> None: ...

global___VideoSourceStats = VideoSourceStats

@typing.final
class AudioPlayoutStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KIND_FIELD_NUMBER: builtins.int
    SYNTHESIZED_SAMPLES_DURATION_FIELD_NUMBER: builtins.int
    SYNTHESIZED_SAMPLES_EVENTS_FIELD_NUMBER: builtins.int
    TOTAL_SAMPLES_DURATION_FIELD_NUMBER: builtins.int
    TOTAL_PLAYOUT_DELAY_FIELD_NUMBER: builtins.int
    TOTAL_SAMPLES_COUNT_FIELD_NUMBER: builtins.int
    kind: builtins.str
    synthesized_samples_duration: builtins.float
    synthesized_samples_events: builtins.int
    total_samples_duration: builtins.float
    total_playout_delay: builtins.float
    total_samples_count: builtins.int
    def __init__(
        self,
        *,
        kind: builtins.str | None = ...,
        synthesized_samples_duration: builtins.float | None = ...,
        synthesized_samples_events: builtins.int | None = ...,
        total_samples_duration: builtins.float | None = ...,
        total_playout_delay: builtins.float | None = ...,
        total_samples_count: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["kind", b"kind", "synthesized_samples_duration", b"synthesized_samples_duration", "synthesized_samples_events", b"synthesized_samples_events", "total_playout_delay", b"total_playout_delay", "total_samples_count", b"total_samples_count", "total_samples_duration", b"total_samples_duration"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["kind", b"kind", "synthesized_samples_duration", b"synthesized_samples_duration", "synthesized_samples_events", b"synthesized_samples_events", "total_playout_delay", b"total_playout_delay", "total_samples_count", b"total_samples_count", "total_samples_duration", b"total_samples_duration"]) -> None: ...

global___AudioPlayoutStats = AudioPlayoutStats

@typing.final
class PeerConnectionStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATA_CHANNELS_OPENED_FIELD_NUMBER: builtins.int
    DATA_CHANNELS_CLOSED_FIELD_NUMBER: builtins.int
    data_channels_opened: builtins.int
    data_channels_closed: builtins.int
    def __init__(
        self,
        *,
        data_channels_opened: builtins.int | None = ...,
        data_channels_closed: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["data_channels_closed", b"data_channels_closed", "data_channels_opened", b"data_channels_opened"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["data_channels_closed", b"data_channels_closed", "data_channels_opened", b"data_channels_opened"]) -> None: ...

global___PeerConnectionStats = PeerConnectionStats

@typing.final
class DataChannelStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LABEL_FIELD_NUMBER: builtins.int
    PROTOCOL_FIELD_NUMBER: builtins.int
    DATA_CHANNEL_IDENTIFIER_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    MESSAGES_SENT_FIELD_NUMBER: builtins.int
    BYTES_SENT_FIELD_NUMBER: builtins.int
    MESSAGES_RECEIVED_FIELD_NUMBER: builtins.int
    BYTES_RECEIVED_FIELD_NUMBER: builtins.int
    label: builtins.str
    protocol: builtins.str
    data_channel_identifier: builtins.int
    state: global___DataChannelState.ValueType
    messages_sent: builtins.int
    bytes_sent: builtins.int
    messages_received: builtins.int
    bytes_received: builtins.int
    def __init__(
        self,
        *,
        label: builtins.str | None = ...,
        protocol: builtins.str | None = ...,
        data_channel_identifier: builtins.int | None = ...,
        state: global___DataChannelState.ValueType | None = ...,
        messages_sent: builtins.int | None = ...,
        bytes_sent: builtins.int | None = ...,
        messages_received: builtins.int | None = ...,
        bytes_received: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bytes_received", b"bytes_received", "bytes_sent", b"bytes_sent", "data_channel_identifier", b"data_channel_identifier", "label", b"label", "messages_received", b"messages_received", "messages_sent", b"messages_sent", "protocol", b"protocol", "state", b"state"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bytes_received", b"bytes_received", "bytes_sent", b"bytes_sent", "data_channel_identifier", b"data_channel_identifier", "label", b"label", "messages_received", b"messages_received", "messages_sent", b"messages_sent", "protocol", b"protocol", "state", b"state"]) -> None: ...

global___DataChannelStats = DataChannelStats

@typing.final
class TransportStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PACKETS_SENT_FIELD_NUMBER: builtins.int
    PACKETS_RECEIVED_FIELD_NUMBER: builtins.int
    BYTES_SENT_FIELD_NUMBER: builtins.int
    BYTES_RECEIVED_FIELD_NUMBER: builtins.int
    ICE_ROLE_FIELD_NUMBER: builtins.int
    ICE_LOCAL_USERNAME_FRAGMENT_FIELD_NUMBER: builtins.int
    DTLS_STATE_FIELD_NUMBER: builtins.int
    ICE_STATE_FIELD_NUMBER: builtins.int
    SELECTED_CANDIDATE_PAIR_ID_FIELD_NUMBER: builtins.int
    LOCAL_CERTIFICATE_ID_FIELD_NUMBER: builtins.int
    REMOTE_CERTIFICATE_ID_FIELD_NUMBER: builtins.int
    TLS_VERSION_FIELD_NUMBER: builtins.int
    DTLS_CIPHER_FIELD_NUMBER: builtins.int
    DTLS_ROLE_FIELD_NUMBER: builtins.int
    SRTP_CIPHER_FIELD_NUMBER: builtins.int
    SELECTED_CANDIDATE_PAIR_CHANGES_FIELD_NUMBER: builtins.int
    packets_sent: builtins.int
    packets_received: builtins.int
    bytes_sent: builtins.int
    bytes_received: builtins.int
    ice_role: global___IceRole.ValueType
    ice_local_username_fragment: builtins.str
    dtls_state: global___DtlsTransportState.ValueType
    ice_state: global___IceTransportState.ValueType
    selected_candidate_pair_id: builtins.str
    local_certificate_id: builtins.str
    remote_certificate_id: builtins.str
    tls_version: builtins.str
    dtls_cipher: builtins.str
    dtls_role: global___DtlsRole.ValueType
    srtp_cipher: builtins.str
    selected_candidate_pair_changes: builtins.int
    def __init__(
        self,
        *,
        packets_sent: builtins.int | None = ...,
        packets_received: builtins.int | None = ...,
        bytes_sent: builtins.int | None = ...,
        bytes_received: builtins.int | None = ...,
        ice_role: global___IceRole.ValueType | None = ...,
        ice_local_username_fragment: builtins.str | None = ...,
        dtls_state: global___DtlsTransportState.ValueType | None = ...,
        ice_state: global___IceTransportState.ValueType | None = ...,
        selected_candidate_pair_id: builtins.str | None = ...,
        local_certificate_id: builtins.str | None = ...,
        remote_certificate_id: builtins.str | None = ...,
        tls_version: builtins.str | None = ...,
        dtls_cipher: builtins.str | None = ...,
        dtls_role: global___DtlsRole.ValueType | None = ...,
        srtp_cipher: builtins.str | None = ...,
        selected_candidate_pair_changes: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bytes_received", b"bytes_received", "bytes_sent", b"bytes_sent", "dtls_cipher", b"dtls_cipher", "dtls_role", b"dtls_role", "dtls_state", b"dtls_state", "ice_local_username_fragment", b"ice_local_username_fragment", "ice_role", b"ice_role", "ice_state", b"ice_state", "local_certificate_id", b"local_certificate_id", "packets_received", b"packets_received", "packets_sent", b"packets_sent", "remote_certificate_id", b"remote_certificate_id", "selected_candidate_pair_changes", b"selected_candidate_pair_changes", "selected_candidate_pair_id", b"selected_candidate_pair_id", "srtp_cipher", b"srtp_cipher", "tls_version", b"tls_version"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bytes_received", b"bytes_received", "bytes_sent", b"bytes_sent", "dtls_cipher", b"dtls_cipher", "dtls_role", b"dtls_role", "dtls_state", b"dtls_state", "ice_local_username_fragment", b"ice_local_username_fragment", "ice_role", b"ice_role", "ice_state", b"ice_state", "local_certificate_id", b"local_certificate_id", "packets_received", b"packets_received", "packets_sent", b"packets_sent", "remote_certificate_id", b"remote_certificate_id", "selected_candidate_pair_changes", b"selected_candidate_pair_changes", "selected_candidate_pair_id", b"selected_candidate_pair_id", "srtp_cipher", b"srtp_cipher", "tls_version", b"tls_version"]) -> None: ...

global___TransportStats = TransportStats

@typing.final
class CandidatePairStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TRANSPORT_ID_FIELD_NUMBER: builtins.int
    LOCAL_CANDIDATE_ID_FIELD_NUMBER: builtins.int
    REMOTE_CANDIDATE_ID_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    NOMINATED_FIELD_NUMBER: builtins.int
    PACKETS_SENT_FIELD_NUMBER: builtins.int
    PACKETS_RECEIVED_FIELD_NUMBER: builtins.int
    BYTES_SENT_FIELD_NUMBER: builtins.int
    BYTES_RECEIVED_FIELD_NUMBER: builtins.int
    LAST_PACKET_SENT_TIMESTAMP_FIELD_NUMBER: builtins.int
    LAST_PACKET_RECEIVED_TIMESTAMP_FIELD_NUMBER: builtins.int
    TOTAL_ROUND_TRIP_TIME_FIELD_NUMBER: builtins.int
    CURRENT_ROUND_TRIP_TIME_FIELD_NUMBER: builtins.int
    AVAILABLE_OUTGOING_BITRATE_FIELD_NUMBER: builtins.int
    AVAILABLE_INCOMING_BITRATE_FIELD_NUMBER: builtins.int
    REQUESTS_RECEIVED_FIELD_NUMBER: builtins.int
    REQUESTS_SENT_FIELD_NUMBER: builtins.int
    RESPONSES_RECEIVED_FIELD_NUMBER: builtins.int
    RESPONSES_SENT_FIELD_NUMBER: builtins.int
    CONSENT_REQUESTS_SENT_FIELD_NUMBER: builtins.int
    PACKETS_DISCARDED_ON_SEND_FIELD_NUMBER: builtins.int
    BYTES_DISCARDED_ON_SEND_FIELD_NUMBER: builtins.int
    transport_id: builtins.str
    local_candidate_id: builtins.str
    remote_candidate_id: builtins.str
    state: global___IceCandidatePairState.ValueType
    nominated: builtins.bool
    packets_sent: builtins.int
    packets_received: builtins.int
    bytes_sent: builtins.int
    bytes_received: builtins.int
    last_packet_sent_timestamp: builtins.float
    last_packet_received_timestamp: builtins.float
    total_round_trip_time: builtins.float
    current_round_trip_time: builtins.float
    available_outgoing_bitrate: builtins.float
    available_incoming_bitrate: builtins.float
    requests_received: builtins.int
    requests_sent: builtins.int
    responses_received: builtins.int
    responses_sent: builtins.int
    consent_requests_sent: builtins.int
    packets_discarded_on_send: builtins.int
    bytes_discarded_on_send: builtins.int
    def __init__(
        self,
        *,
        transport_id: builtins.str | None = ...,
        local_candidate_id: builtins.str | None = ...,
        remote_candidate_id: builtins.str | None = ...,
        state: global___IceCandidatePairState.ValueType | None = ...,
        nominated: builtins.bool | None = ...,
        packets_sent: builtins.int | None = ...,
        packets_received: builtins.int | None = ...,
        bytes_sent: builtins.int | None = ...,
        bytes_received: builtins.int | None = ...,
        last_packet_sent_timestamp: builtins.float | None = ...,
        last_packet_received_timestamp: builtins.float | None = ...,
        total_round_trip_time: builtins.float | None = ...,
        current_round_trip_time: builtins.float | None = ...,
        available_outgoing_bitrate: builtins.float | None = ...,
        available_incoming_bitrate: builtins.float | None = ...,
        requests_received: builtins.int | None = ...,
        requests_sent: builtins.int | None = ...,
        responses_received: builtins.int | None = ...,
        responses_sent: builtins.int | None = ...,
        consent_requests_sent: builtins.int | None = ...,
        packets_discarded_on_send: builtins.int | None = ...,
        bytes_discarded_on_send: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["available_incoming_bitrate", b"available_incoming_bitrate", "available_outgoing_bitrate", b"available_outgoing_bitrate", "bytes_discarded_on_send", b"bytes_discarded_on_send", "bytes_received", b"bytes_received", "bytes_sent", b"bytes_sent", "consent_requests_sent", b"consent_requests_sent", "current_round_trip_time", b"current_round_trip_time", "last_packet_received_timestamp", b"last_packet_received_timestamp", "last_packet_sent_timestamp", b"last_packet_sent_timestamp", "local_candidate_id", b"local_candidate_id", "nominated", b"nominated", "packets_discarded_on_send", b"packets_discarded_on_send", "packets_received", b"packets_received", "packets_sent", b"packets_sent", "remote_candidate_id", b"remote_candidate_id", "requests_received", b"requests_received", "requests_sent", b"requests_sent", "responses_received", b"responses_received", "responses_sent", b"responses_sent", "state", b"state", "total_round_trip_time", b"total_round_trip_time", "transport_id", b"transport_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["available_incoming_bitrate", b"available_incoming_bitrate", "available_outgoing_bitrate", b"available_outgoing_bitrate", "bytes_discarded_on_send", b"bytes_discarded_on_send", "bytes_received", b"bytes_received", "bytes_sent", b"bytes_sent", "consent_requests_sent", b"consent_requests_sent", "current_round_trip_time", b"current_round_trip_time", "last_packet_received_timestamp", b"last_packet_received_timestamp", "last_packet_sent_timestamp", b"last_packet_sent_timestamp", "local_candidate_id", b"local_candidate_id", "nominated", b"nominated", "packets_discarded_on_send", b"packets_discarded_on_send", "packets_received", b"packets_received", "packets_sent", b"packets_sent", "remote_candidate_id", b"remote_candidate_id", "requests_received", b"requests_received", "requests_sent", b"requests_sent", "responses_received", b"responses_received", "responses_sent", b"responses_sent", "state", b"state", "total_round_trip_time", b"total_round_trip_time", "transport_id", b"transport_id"]) -> None: ...

global___CandidatePairStats = CandidatePairStats

@typing.final
class IceCandidateStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TRANSPORT_ID_FIELD_NUMBER: builtins.int
    ADDRESS_FIELD_NUMBER: builtins.int
    PORT_FIELD_NUMBER: builtins.int
    PROTOCOL_FIELD_NUMBER: builtins.int
    CANDIDATE_TYPE_FIELD_NUMBER: builtins.int
    PRIORITY_FIELD_NUMBER: builtins.int
    URL_FIELD_NUMBER: builtins.int
    RELAY_PROTOCOL_FIELD_NUMBER: builtins.int
    FOUNDATION_FIELD_NUMBER: builtins.int
    RELATED_ADDRESS_FIELD_NUMBER: builtins.int
    RELATED_PORT_FIELD_NUMBER: builtins.int
    USERNAME_FRAGMENT_FIELD_NUMBER: builtins.int
    TCP_TYPE_FIELD_NUMBER: builtins.int
    transport_id: builtins.str
    address: builtins.str
    port: builtins.int
    protocol: builtins.str
    candidate_type: global___IceCandidateType.ValueType
    priority: builtins.int
    url: builtins.str
    relay_protocol: global___IceServerTransportProtocol.ValueType
    foundation: builtins.str
    related_address: builtins.str
    related_port: builtins.int
    username_fragment: builtins.str
    tcp_type: global___IceTcpCandidateType.ValueType
    def __init__(
        self,
        *,
        transport_id: builtins.str | None = ...,
        address: builtins.str | None = ...,
        port: builtins.int | None = ...,
        protocol: builtins.str | None = ...,
        candidate_type: global___IceCandidateType.ValueType | None = ...,
        priority: builtins.int | None = ...,
        url: builtins.str | None = ...,
        relay_protocol: global___IceServerTransportProtocol.ValueType | None = ...,
        foundation: builtins.str | None = ...,
        related_address: builtins.str | None = ...,
        related_port: builtins.int | None = ...,
        username_fragment: builtins.str | None = ...,
        tcp_type: global___IceTcpCandidateType.ValueType | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["address", b"address", "candidate_type", b"candidate_type", "foundation", b"foundation", "port", b"port", "priority", b"priority", "protocol", b"protocol", "related_address", b"related_address", "related_port", b"related_port", "relay_protocol", b"relay_protocol", "tcp_type", b"tcp_type", "transport_id", b"transport_id", "url", b"url", "username_fragment", b"username_fragment"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["address", b"address", "candidate_type", b"candidate_type", "foundation", b"foundation", "port", b"port", "priority", b"priority", "protocol", b"protocol", "related_address", b"related_address", "related_port", b"related_port", "relay_protocol", b"relay_protocol", "tcp_type", b"tcp_type", "transport_id", b"transport_id", "url", b"url", "username_fragment", b"username_fragment"]) -> None: ...

global___IceCandidateStats = IceCandidateStats

@typing.final
class CertificateStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FINGERPRINT_FIELD_NUMBER: builtins.int
    FINGERPRINT_ALGORITHM_FIELD_NUMBER: builtins.int
    BASE64_CERTIFICATE_FIELD_NUMBER: builtins.int
    ISSUER_CERTIFICATE_ID_FIELD_NUMBER: builtins.int
    fingerprint: builtins.str
    fingerprint_algorithm: builtins.str
    base64_certificate: builtins.str
    issuer_certificate_id: builtins.str
    def __init__(
        self,
        *,
        fingerprint: builtins.str | None = ...,
        fingerprint_algorithm: builtins.str | None = ...,
        base64_certificate: builtins.str | None = ...,
        issuer_certificate_id: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["base64_certificate", b"base64_certificate", "fingerprint", b"fingerprint", "fingerprint_algorithm", b"fingerprint_algorithm", "issuer_certificate_id", b"issuer_certificate_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["base64_certificate", b"base64_certificate", "fingerprint", b"fingerprint", "fingerprint_algorithm", b"fingerprint_algorithm", "issuer_certificate_id", b"issuer_certificate_id"]) -> None: ...

global___CertificateStats = CertificateStats

@typing.final
class StreamStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    STREAM_IDENTIFIER_FIELD_NUMBER: builtins.int
    id: builtins.str
    stream_identifier: builtins.str
    """required int64 timestamp = 3;"""
    def __init__(
        self,
        *,
        id: builtins.str | None = ...,
        stream_identifier: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["id", b"id", "stream_identifier", b"stream_identifier"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["id", b"id", "stream_identifier", b"stream_identifier"]) -> None: ...

global___StreamStats = StreamStats



from livekit.rtc import AudioResampler, AudioResamplerQuality
import time
import wave
import os


def test_audio_resampler():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    wav_file_path = os.path.join(current_dir, "test_audio.wav")

    # Open the wave file
    with wave.open(wav_file_path, "rb") as wf_in:
        n_channels = wf_in.getnchannels()
        sampwidth = wf_in.getsampwidth()
        n_frames = wf_in.getnframes()
        audio_data = bytearray(wf_in.readframes(n_frames))

    if sampwidth != 2:
        raise ValueError(f"Expected 16-bit PCM data, but got {sampwidth * 8}-bit.")

    qualities = [
        AudioResamplerQuality.QUICK,
        AudioResamplerQuality.LOW,
        AudioResamplerQuality.MEDIUM,
        AudioResamplerQuality.HIGH,
        AudioResamplerQuality.VERY_HIGH,
    ]

    for quality in qualities:
        total_time = 0
        nb_runs = 20
        output_frames = []
        for i in range(nb_runs):
            output_frames = []
            resampler = AudioResampler(44100, 8000, quality=quality)

            start_time = time.perf_counter()

            chunk_size = 1024 * n_channels * sampwidth
            output_frames = []
            for i in range(0, len(audio_data), chunk_size):
                chunk = audio_data[i : i + chunk_size]
                frames = resampler.push(bytearray(chunk))
                output_frames.extend(frames)

            frames = resampler.flush()
            output_frames.extend(frames)

            end_time = time.perf_counter()
            total_time += end_time - start_time

        total_time = total_time * 1000
        print(f"Quality: {quality}, Average time: {total_time / nb_runs:.2f}ms")

        output_data = b""

        for frame in output_frames:
            output_data += frame.data

        with wave.open(f"audio_resampled_{quality.name}.wav", "wb") as wf_out:
            wf_out.setnchannels(n_channels)
            wf_out.setsampwidth(sampwidth)
            wf_out.setframerate(8000)
            wf_out.writeframes(output_data)



import asyncio
import logging
import time
from collections import deque
from typing import Optional, Union

from .video_frame import VideoFrame
from .audio_frame import AudioFrame
from .audio_source import AudioSource
from .video_source import VideoSource


logger = logging.getLogger(__name__)


class AVSynchronizer:
    """Synchronize audio and video capture.

    Usage:
        av_sync = AVSynchronizer(
            audio_source=audio_source,
            video_source=video_source,
            video_fps=video_fps,
        )

        async for video_frame, audio_frame in video_generator:
            await av_sync.push(video_frame)
            await av_sync.push(audio_frame)
    """

    def __init__(
        self,
        *,
        audio_source: AudioSource,
        video_source: VideoSource,
        video_fps: float,
        video_queue_size_ms: float = 100,
        _max_delay_tolerance_ms: float = 300,
    ):
        self._audio_source = audio_source
        self._video_source = video_source
        self._video_fps = video_fps
        self._video_queue_size_ms = video_queue_size_ms
        self._max_delay_tolerance_ms = _max_delay_tolerance_ms

        self._stopped = False
        # the time of the last video/audio frame captured
        self._last_video_time: float = 0
        self._last_audio_time: float = 0

        self._video_queue_max_size = int(self._video_fps * self._video_queue_size_ms / 1000)
        if self._video_queue_size_ms > 0:
            # ensure queue is bounded if queue size is specified
            self._video_queue_max_size = max(1, self._video_queue_max_size)

        self._video_queue = asyncio.Queue[tuple[VideoFrame, Optional[float]]](
            maxsize=self._video_queue_max_size
        )
        self._fps_controller = _FPSController(
            expected_fps=self._video_fps,
            max_delay_tolerance_ms=self._max_delay_tolerance_ms,
        )
        self._capture_video_task = asyncio.create_task(self._capture_video())

    async def push(
        self, frame: Union[VideoFrame, AudioFrame], timestamp: Optional[float] = None
    ) -> None:
        """Push a frame to the synchronizer

        Args:
            frame: The video or audio frame to push.
            timestamp: (optional) The timestamp of the frame, for logging purposes for now.
                For AudioFrame, it should be the end time of the frame.
        """
        if isinstance(frame, AudioFrame):
            await self._audio_source.capture_frame(frame)
            if timestamp is not None:
                self._last_audio_time = timestamp
            return

        await self._video_queue.put((frame, timestamp))

    async def clear_queue(self) -> None:
        self._audio_source.clear_queue()
        while not self._video_queue.empty():
            await self._video_queue.get()
            self._video_queue.task_done()

    async def wait_for_playout(self) -> None:
        """Wait until all video and audio frames are played out."""
        await asyncio.gather(
            self._audio_source.wait_for_playout(),
            self._video_queue.join(),
        )

    def reset(self) -> None:
        self._fps_controller.reset()

    async def _capture_video(self) -> None:
        while not self._stopped:
            frame, timestamp = await self._video_queue.get()
            async with self._fps_controller:
                self._video_source.capture_frame(frame)
                if timestamp is not None:
                    self._last_video_time = timestamp
            self._video_queue.task_done()

    async def aclose(self) -> None:
        self._stopped = True
        if self._capture_video_task:
            self._capture_video_task.cancel()

    @property
    def actual_fps(self) -> float:
        return self._fps_controller.actual_fps

    @property
    def last_video_time(self) -> float:
        """The time of the last video frame captured"""
        return self._last_video_time

    @property
    def last_audio_time(self) -> float:
        """The time of the last audio frame played out"""
        return self._last_audio_time - self._audio_source.queued_duration


class _FPSController:
    def __init__(self, *, expected_fps: float, max_delay_tolerance_ms: float = 300) -> None:
        """Controls frame rate by adjusting sleep time based on actual FPS.

        Usage:
            async with _FPSController(expected_fps=30):
                # process frame
                pass

        Args:
            expected_fps: Target frames per second
            max_delay_tolerance_ms: Maximum delay tolerance in milliseconds
        """
        self._expected_fps = expected_fps
        self._frame_interval = 1.0 / expected_fps
        self._max_delay_tolerance_secs = max_delay_tolerance_ms / 1000

        self._next_frame_time: Optional[float] = None
        self._fps_calc_winsize = max(2, int(1.0 * expected_fps))
        self._send_timestamps: deque[float] = deque(maxlen=self._fps_calc_winsize)

    async def __aenter__(self) -> None:
        await self.wait_next_process()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self.after_process()

    def reset(self) -> None:
        self._next_frame_time = None
        self._send_timestamps.clear()

    async def wait_next_process(self) -> None:
        """Wait until it's time for the next frame.

        Adjusts sleep time based on actual FPS to maintain target rate.
        """
        current_time = time.perf_counter()

        # initialize the next frame time
        if self._next_frame_time is None:
            self._next_frame_time = current_time

        # calculate sleep time
        sleep_time = self._next_frame_time - current_time
        if sleep_time > 0:
            await asyncio.sleep(sleep_time)
        else:
            # check if significantly behind schedule
            if -sleep_time > self._max_delay_tolerance_secs:
                logger.warning(f"Frame capture was behind schedule for {-sleep_time * 1000:.2f} ms")
                self._next_frame_time = time.perf_counter()

    def after_process(self) -> None:
        """Update timing information after processing a frame."""
        assert self._next_frame_time is not None, "wait_next_process must be called first"

        # update timing information
        self._send_timestamps.append(time.perf_counter())

        # calculate next frame time
        self._next_frame_time += self._frame_interval

    @property
    def expected_fps(self) -> float:
        return self._expected_fps

    @property
    def actual_fps(self) -> float:
        """Get current average FPS."""
        if len(self._send_timestamps) < 2:
            return 0

        return (len(self._send_timestamps) - 1) / (
            self._send_timestamps[-1] - self._send_timestamps[0]
        )



# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import datetime
import asyncio
import ctypes
import logging
from dataclasses import dataclass, field
from typing import Callable, Dict, Literal, Optional, cast, Mapping
import warnings

from .event_emitter import EventEmitter
from ._ffi_client import FfiClient, FfiHandle
from ._proto import ffi_pb2 as proto_ffi
from ._proto import participant_pb2 as proto_participant
from ._proto import room_pb2 as proto_room
from ._proto import stats_pb2 as proto_stats
from ._proto.room_pb2 import ConnectionState
from ._proto.track_pb2 import TrackKind
from ._proto.rpc_pb2 import RpcMethodInvocationEvent
from ._utils import BroadcastQueue
from .e2ee import E2EEManager, E2EEOptions
from .participant import LocalParticipant, Participant, RemoteParticipant
from .track import RemoteAudioTrack, RemoteVideoTrack
from .track_publication import RemoteTrackPublication, TrackPublication
from .transcription import TranscriptionSegment
from .data_stream import (
    TextStreamReader,
    ByteStreamReader,
    TextStreamHandler,
    ByteStreamHandler,
)


EventTypes = Literal[
    "participant_connected",
    "participant_disconnected",
    "local_track_published",
    "local_track_unpublished",
    "local_track_subscribed",
    "track_published",
    "track_unpublished",
    "track_subscribed",
    "track_unsubscribed",
    "track_subscription_failed",
    "track_muted",
    "track_unmuted",
    "active_speakers_changed",
    "room_metadata_changed",
    "participant_metadata_changed",
    "participant_name_changed",
    "participant_attributes_changed",
    "connection_quality_changed",
    "participant_encryption_status_changed",
    "data_received",
    "sip_dtmf_received",
    "transcription_received",
    "e2ee_state_changed",
    "connection_state_changed",
    "connected",
    "disconnected",
    "reconnecting",
    "reconnected",
    "room_updated",
    "moved",
]


@dataclass
class RtcConfiguration:
    ice_transport_type: proto_room.IceTransportType.ValueType = (
        proto_room.IceTransportType.TRANSPORT_ALL
    )
    """Specifies the type of ICE transport to be used (e.g., all, relay, etc.)."""
    continual_gathering_policy: proto_room.ContinualGatheringPolicy.ValueType = (
        proto_room.ContinualGatheringPolicy.GATHER_CONTINUALLY
    )
    """Policy for continual gathering of ICE candidates."""
    ice_servers: list[proto_room.IceServer] = field(default_factory=list)
    """List of ICE servers for STUN/TURN. When empty, it uses the default ICE servers provided by
    the SFU."""


@dataclass
class RoomOptions:
    auto_subscribe: bool = True
    """Automatically subscribe to tracks when participants join."""
    dynacast: bool = False
    e2ee: E2EEOptions | None = None
    """Deprecated, use `encryption` field instead"""
    encryption: E2EEOptions | None = None
    """Options for end-to-end encryption."""
    rtc_config: RtcConfiguration | None = None
    """WebRTC-related configuration."""


@dataclass
class DataPacket:
    data: bytes
    """The payload of the data packet."""
    kind: proto_room.DataPacketKind.ValueType
    """Type of the data packet (e.g., RELIABLE, LOSSY)."""
    participant: RemoteParticipant | None
    """Participant who sent the data. None when sent by a server SDK."""
    topic: str | None = None
    """Topic associated with the data packet."""


@dataclass
class SipDTMF:
    code: int
    """DTMF code corresponding to the digit."""
    digit: str
    """DTMF digit sent."""
    participant: RemoteParticipant | None = None
    """Participant who sent the DTMF digit. None when sent by a server SDK."""


@dataclass
class RtcStats:
    publisher_stats: list[proto_stats.RtcStats]
    subscriber_stats: list[proto_stats.RtcStats]


class ConnectError(Exception):
    def __init__(self, message: str):
        self.message = message


class Room(EventEmitter[EventTypes]):
    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        """Initializes a new Room instance.

        Parameters:
            loop (Optional[asyncio.AbstractEventLoop]): The event loop to use. If not provided, the default event loop is used.
        """
        super().__init__()

        self._ffi_handle: Optional[FfiHandle] = None
        self._loop = loop or asyncio.get_event_loop()
        self._room_queue = BroadcastQueue[proto_ffi.FfiEvent]()
        self._info = proto_room.RoomInfo()
        self._rpc_invocation_tasks: set[asyncio.Task] = set()
        self._data_stream_tasks: set[asyncio.Task] = set()

        self._remote_participants: Dict[str, RemoteParticipant] = {}
        self._connection_state = ConnectionState.CONN_DISCONNECTED
        self._first_sid_future = asyncio.Future[str]()
        self._local_participant: LocalParticipant | None = None

        self._text_stream_readers: Dict[str, TextStreamReader] = {}
        self._byte_stream_readers: Dict[str, ByteStreamReader] = {}
        self._text_stream_handlers: Dict[str, TextStreamHandler] = {}
        self._byte_stream_handlers: Dict[str, ByteStreamHandler] = {}

    def __del__(self) -> None:
        if self._ffi_handle is not None:
            FfiClient.instance.queue.unsubscribe(self._ffi_queue)

    @property
    async def sid(self) -> str:
        """Asynchronously retrieves the session ID (SID) of the room.

        Returns:
            str: The session ID of the room.
        """
        if self._info.sid:
            return self._info.sid

        return await self._first_sid_future

    @property
    def local_participant(self) -> LocalParticipant:
        """Gets the local participant in the room.

        Returns:
            LocalParticipant: The local participant in the room.
        """
        if self._local_participant is None:
            raise Exception("cannot access local participant before connecting")

        return self._local_participant

    @property
    def connection_state(self) -> ConnectionState.ValueType:
        """Gets the connection state of the room.

        Returns:
            ConnectionState: The connection state of the room.
        """
        return self._connection_state

    @property
    def remote_participants(self) -> Mapping[str, RemoteParticipant]:
        """Gets the remote participants in the room.

        Returns:
            dict[str, RemoteParticipant]: A dictionary of remote participants indexed by their
            identity.
        """
        return self._remote_participants

    @property
    def name(self) -> str:
        """Gets the name of the room.

        Returns:
            str: The name of the room.
        """
        return self._info.name

    @property
    def metadata(self) -> str:
        """Gets the metadata associated with the room.

        Returns:
            str: The metadata of the room.
        """
        return self._info.metadata

    @property
    def e2ee_manager(self) -> E2EEManager:
        """Gets the end-to-end encryption (E2EE) manager for the room.

        Returns:
            E2EEManager: The E2EE manager instance.
        """
        return self._e2ee_manager

    @property
    def num_participants(self) -> int:
        """Gets the number of participants in the room.
        This value is updated periodically, and is eventually consistent.

        Returns:
            int: The number of participants in the room.
        """
        return self._info.num_participants

    @property
    def num_publishers(self) -> int:
        """Gets the number of publishers in the room.
        This value is updated periodically, and is eventually consistent.

        Returns:
            int: The number of publishers in the room.
        """
        return self._info.num_publishers

    @property
    def creation_time(self) -> datetime.datetime:
        """Time when the room was created.

        Returns:
            datetime.datetime: The creation time of the room.
        """
        return datetime.datetime.fromtimestamp(
            self._info.creation_time / 1000, datetime.timezone.utc
        )

    @property
    def is_recording(self) -> bool:
        """Whether the room is actively recording.

        Returns:
            bool: True if actively recording, False otherwise.
        """
        return self._info.active_recording

    @property
    def departure_timeout(self) -> float:
        """Amount of time to hold the room open after the last standard participant leaves.

        Returns:
            float: The departure timeout of the room.
        """
        return float(self._info.departure_timeout)

    @property
    def empty_timeout(self) -> float:
        """Amount of time to keep the room open if no participants join.

        Returns:
            float: The empty timeout of the room.
        """
        return float(self._info.empty_timeout)

    def isconnected(self) -> bool:
        """Checks if the room is currently connected.

        Returns:
            bool: True if connected, False otherwise.
        """
        return (
            self._ffi_handle is not None
            and self._connection_state != ConnectionState.CONN_DISCONNECTED
        )

    def on(self, event: EventTypes, callback: Optional[Callable] = None) -> Callable:
        """Registers an event handler for a specific event type.

        Parameters:
            event (EventTypes): The name of the event to listen for.
            callback (Callable): The function to call when the event occurs.

        Returns:
            Callable: The registered callback function.

        Available events:
            - **"participant_connected"**: Called when a new participant joins the room.
                - Arguments: `participant` (RemoteParticipant)
            - **"participant_disconnected"**: Called when a participant leaves the room.
                - Arguments: `participant` (RemoteParticipant)
            - **"local_track_published"**: Called when a local track is published.
                - Arguments: `publication` (LocalTrackPublication), `track` (Track)
            - **"local_track_unpublished"**: Called when a local track is unpublished.
                - Arguments: `publication` (LocalTrackPublication)
            - **"local_track_subscribed"**: Called when a local track is subscribed.
                - Arguments: `track` (Track)
            - **"track_published"**: Called when a remote participant publishes a track.
                - Arguments: `publication` (RemoteTrackPublication), `participant` (RemoteParticipant)
            - **"track_unpublished"**: Called when a remote participant unpublishes a track.
                - Arguments: `publication` (RemoteTrackPublication), `participant` (RemoteParticipant)
            - **"track_subscribed"**: Called when a track is subscribed.
                - Arguments: `track` (Track), `publication` (RemoteTrackPublication), `participant` (RemoteParticipant)
            - **"track_unsubscribed"**: Called when a track is unsubscribed.
                - Arguments: `track` (Track), `publication` (RemoteTrackPublication), `participant` (RemoteParticipant)
            - **"track_subscription_failed"**: Called when a track subscription fails.
                - Arguments: `participant` (RemoteParticipant), `track_sid` (str), `error` (str)
            - **"track_muted"**: Called when a track is muted.
                - Arguments: `participant` (Participant), `publication` (TrackPublication)
            - **"track_unmuted"**: Called when a track is unmuted.
                - Arguments: `participant` (Participant), `publication` (TrackPublication)
            - **"active_speakers_changed"**: Called when the list of active speakers changes.
                - Arguments: `speakers` (list[Participant])
            - **"room_metadata_changed"**: Called when the room's metadata is updated.
                - Arguments: `old_metadata` (str), `new_metadata` (str)
            - **"participant_metadata_changed"**: Called when a participant's metadata is updated.
                - Arguments: `participant` (Participant), `old_metadata` (str), `new_metadata` (str)
            - **"participant_name_changed"**: Called when a participant's name is changed.
                - Arguments: `participant` (Participant), `old_name` (str), `new_name` (str)
            - **"participant_attributes_changed"**: Called when a participant's attributes change.
                - Arguments: `changed_attributes` (dict), `participant` (Participant)
            - **"participant_encryption_status_changed"**: Called when a participant's encryption status changes.
                - Arguments `is_encrypted` (bool), `participant` (Participant)
            - **"connection_quality_changed"**: Called when a participant's connection quality changes.
                - Arguments: `participant` (Participant), `quality` (ConnectionQuality)
            - **"transcription_received"**: Called when a transcription is received.
                - Arguments: `segments` (list[TranscriptionSegment]), `participant` (Participant), `publication` (TrackPublication)
            - **"data_received"**: Called when data is received.
                - Arguments: `data_packet` (DataPacket)
            - **"sip_dtmf_received"**: Called when a SIP DTMF signal is received.
                - Arguments: `sip_dtmf` (SipDTMF)
            - **"e2ee_state_changed"**: Called when a participant's E2EE state changes.
                - Arguments: `participant` (Participant), `state` (EncryptionState)
            - **"connection_state_changed"**: Called when the room's connection state changes.
                - Arguments: `connection_state` (ConnectionState)
            - **"connected"**: Called when the room is successfully connected.
                - Arguments: None
            - **"disconnected"**: Called when the room is disconnected.
                - Arguments: `reason` (DisconnectReason)
            - **"reconnecting"**: Called when the room is attempting to reconnect.
                - Arguments: None
            - **"reconnected"**: Called when the room has successfully reconnected.
                - Arguments: None
            - **"room_updated"**: Called when any information about the room is updated.
                - Arguments: None
            - **"moved"**: Called when the participant has been moved to another room.
                - Arguments: None

        Example:
            ```python
            def on_participant_connected(participant):
                print(f"Participant connected: {participant.identity}")

            room.on("participant_connected", on_participant_connected)
            ```
        """
        return super().on(event, callback)

    async def connect(self, url: str, token: str, options: RoomOptions = RoomOptions()) -> None:
        """Connects to a LiveKit room using the specified URL and token.

        Parameters:
            url (str): The WebSocket URL of the LiveKit server to connect to.
            token (str): The access token for authentication and authorization.
            options (RoomOptions, optional): Additional options for the room connection.

        Raises:
            ConnectError: If the connection fails.

        Example:
            ```python
            room = Room()

            # Listen for events before connecting to the room
            @room.on("participant_connected")
            def on_participant_connected(participant):
                print(f"Participant connected: {participant.identity}")

            await room.connect("ws://localhost:7880", "your_token")
            ```
        """
        req = proto_ffi.FfiRequest()
        req.connect.url = url
        req.connect.token = token

        # options
        req.connect.options.auto_subscribe = options.auto_subscribe
        req.connect.options.dynacast = options.dynacast

        if options.e2ee:
            warnings.warn(
                "options.e2ee is deprecated, use options.encryption instead",
                DeprecationWarning,
                stacklevel=2,
            )

            req.connect.options.e2ee.encryption_type = options.e2ee.encryption_type
            req.connect.options.e2ee.key_provider_options.shared_key = (
                options.e2ee.key_provider_options.shared_key  # type: ignore
            )
            req.connect.options.e2ee.key_provider_options.ratchet_salt = (
                options.e2ee.key_provider_options.ratchet_salt
            )
            req.connect.options.e2ee.key_provider_options.failure_tolerance = (
                options.e2ee.key_provider_options.failure_tolerance
            )
            req.connect.options.e2ee.key_provider_options.ratchet_window_size = (
                options.e2ee.key_provider_options.ratchet_window_size
            )

        if options.encryption:
            req.connect.options.encryption.encryption_type = options.encryption.encryption_type
            req.connect.options.encryption.key_provider_options.shared_key = (
                options.encryption.key_provider_options.shared_key  # type: ignore
            )
            req.connect.options.encryption.key_provider_options.ratchet_salt = (
                options.encryption.key_provider_options.ratchet_salt
            )
            req.connect.options.encryption.key_provider_options.failure_tolerance = (
                options.encryption.key_provider_options.failure_tolerance
            )
            req.connect.options.encryption.key_provider_options.ratchet_window_size = (
                options.encryption.key_provider_options.ratchet_window_size
            )

        if options.rtc_config:
            req.connect.options.rtc_config.ice_transport_type = (
                options.rtc_config.ice_transport_type
            )  # type: ignore
            req.connect.options.rtc_config.continual_gathering_policy = (
                options.rtc_config.continual_gathering_policy
            )  # type: ignore
            req.connect.options.rtc_config.ice_servers.extend(options.rtc_config.ice_servers)

        # subscribe before connecting so we don't miss any events
        self._ffi_queue = FfiClient.instance.queue.subscribe(self._loop)

        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            cb: proto_ffi.FfiEvent = await queue.wait_for(
                lambda e: e.connect.async_id == resp.connect.async_id
            )
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

        if cb.connect.error:
            FfiClient.instance.queue.unsubscribe(self._ffi_queue)
            raise ConnectError(cb.connect.error)

        self._ffi_handle = FfiHandle(cb.connect.result.room.handle.id)

        self._e2ee_manager = E2EEManager(
            self._ffi_handle.handle, options.encryption or options.e2ee
        )

        self._info = cb.connect.result.room.info
        self._connection_state = ConnectionState.CONN_CONNECTED

        self._local_participant = LocalParticipant(
            self._room_queue, cb.connect.result.local_participant
        )

        for pt in cb.connect.result.participants:
            rp = self._create_remote_participant(pt.participant)

            # add the initial remote participant tracks
            for owned_publication_info in pt.publications:
                publication = RemoteTrackPublication(owned_publication_info)
                rp._track_publications[publication.sid] = publication

        # start listening to room events
        self._task = self._loop.create_task(self._listen_task())

    async def get_rtc_stats(self) -> RtcStats:
        if not self.isconnected():
            raise RuntimeError("the room isn't connected")

        req = proto_ffi.FfiRequest()
        req.get_session_stats.room_handle = self._ffi_handle.handle  # type: ignore

        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            cb: proto_ffi.FfiEvent = await queue.wait_for(
                lambda e: e.get_session_stats.async_id == resp.get_session_stats.async_id
            )
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

        if cb.get_session_stats.error:
            raise RuntimeError(cb.get_session_stats.error)

        publisher_stats = list(cb.get_session_stats.result.publisher_stats)
        subscriber_stats = list(cb.get_session_stats.result.subscriber_stats)

        return RtcStats(publisher_stats=publisher_stats, subscriber_stats=subscriber_stats)

    def register_byte_stream_handler(self, topic: str, handler: ByteStreamHandler):
        existing_handler = self._byte_stream_handlers.get(topic)
        if existing_handler is None:
            self._byte_stream_handlers[topic] = handler
        else:
            raise ValueError("byte stream handler for topic '%s' already set" % topic)

    def unregister_byte_stream_handler(self, topic: str):
        if self._byte_stream_handlers.get(topic):
            self._byte_stream_handlers.pop(topic)

    def register_text_stream_handler(self, topic: str, handler: TextStreamHandler):
        existing_handler = self._text_stream_handlers.get(topic)
        if existing_handler is None:
            self._text_stream_handlers[topic] = handler
        else:
            raise ValueError("text stream handler for topic '%s' already set" % topic)

    def unregister_text_stream_handler(self, topic: str):
        if self._text_stream_handlers.get(topic):
            self._text_stream_handlers.pop(topic)

    async def disconnect(self) -> None:
        """Disconnects from the room."""
        if not self.isconnected():
            return

        await self._drain_rpc_invocation_tasks()
        await self._drain_data_stream_tasks()

        req = proto_ffi.FfiRequest()
        req.disconnect.room_handle = self._ffi_handle.handle  # type: ignore
        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            await queue.wait_for(lambda e: e.disconnect.async_id == resp.disconnect.async_id)
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

        await self._task
        FfiClient.instance.queue.unsubscribe(self._ffi_queue)

    async def _listen_task(self) -> None:
        # listen to incoming room events
        while True:
            event = await self._ffi_queue.get()
            if event.WhichOneof("message") == "rpc_method_invocation":
                self._on_rpc_method_invocation(event.rpc_method_invocation)
            elif event.room_event.room_handle == self._ffi_handle.handle:  # type: ignore
                if event.room_event.HasField("eos"):
                    break

                try:
                    self._on_room_event(event.room_event)
                except Exception:
                    logging.exception(
                        "error running user callback for %s: %s",
                        event.room_event.WhichOneof("message"),
                        event.room_event,
                    )

            # wait for the subscribers to process the event
            # before processing the next one
            self._room_queue.put_nowait(event)
            await self._room_queue.join()

        # Clean up any pending RPC invocation tasks
        await self._drain_rpc_invocation_tasks()
        await self._drain_data_stream_tasks()

    def _on_rpc_method_invocation(self, rpc_invocation: RpcMethodInvocationEvent):
        if self._local_participant is None:
            return

        if rpc_invocation.local_participant_handle == self._local_participant._ffi_handle.handle:
            task = self._loop.create_task(
                self._local_participant._handle_rpc_method_invocation(
                    rpc_invocation.invocation_id,
                    rpc_invocation.method,
                    rpc_invocation.request_id,
                    rpc_invocation.caller_identity,
                    rpc_invocation.payload,
                    rpc_invocation.response_timeout_ms / 1000.0,
                )
            )
            self._rpc_invocation_tasks.add(task)
            task.add_done_callback(self._rpc_invocation_tasks.discard)

    def _on_room_event(self, event: proto_room.RoomEvent):
        which = event.WhichOneof("message")
        if which == "participant_connected":
            rparticipant = self._create_remote_participant(event.participant_connected.info)
            self.emit("participant_connected", rparticipant)
        elif which == "participant_disconnected":
            identity = event.participant_disconnected.participant_identity
            rparticipant = self._remote_participants.pop(identity)
            rparticipant._info.disconnect_reason = event.participant_disconnected.disconnect_reason
            self.emit("participant_disconnected", rparticipant)
        elif which == "local_track_published":
            sid = event.local_track_published.track_sid
            lpublication = self.local_participant.track_publications[sid]
            ltrack = lpublication.track
            self.emit("local_track_published", lpublication, ltrack)
        elif which == "local_track_unpublished":
            sid = event.local_track_unpublished.publication_sid
            lpublication = self.local_participant.track_publications[sid]
            self.emit("local_track_unpublished", lpublication)
        elif which == "local_track_subscribed":
            sid = event.local_track_subscribed.track_sid
            lpublication = self.local_participant.track_publications[sid]
            lpublication._first_subscription.set_result(None)
            self.emit("local_track_subscribed", lpublication.track)
        elif which == "track_published":
            rparticipant = self._remote_participants[event.track_published.participant_identity]
            rpublication = RemoteTrackPublication(event.track_published.publication)
            rparticipant._track_publications[rpublication.sid] = rpublication
            self.emit("track_published", rpublication, rparticipant)
        elif which == "track_unpublished":
            rparticipant = self._remote_participants[event.track_unpublished.participant_identity]
            rpublication = rparticipant._track_publications.pop(
                event.track_unpublished.publication_sid
            )
            self.emit("track_unpublished", rpublication, rparticipant)
        elif which == "track_subscribed":
            owned_track_info = event.track_subscribed.track
            track_info = owned_track_info.info
            rparticipant = self._remote_participants[event.track_subscribed.participant_identity]
            rpublication = rparticipant.track_publications[track_info.sid]
            rpublication._subscribed = True
            if track_info.kind == TrackKind.KIND_VIDEO:
                remote_video_track = RemoteVideoTrack(owned_track_info)
                rpublication._track = remote_video_track
                self.emit("track_subscribed", remote_video_track, rpublication, rparticipant)
            elif track_info.kind == TrackKind.KIND_AUDIO:
                remote_audio_track = RemoteAudioTrack(owned_track_info)
                rpublication._track = remote_audio_track
                self.emit("track_subscribed", remote_audio_track, rpublication, rparticipant)
        elif which == "track_unsubscribed":
            identity = event.track_unsubscribed.participant_identity
            rparticipant = self._remote_participants[identity]
            rpublication = rparticipant.track_publications[event.track_unsubscribed.track_sid]
            rtrack = rpublication.track
            rpublication._track = None
            rpublication._subscribed = False
            self.emit("track_unsubscribed", rtrack, rpublication, rparticipant)
        elif which == "track_subscription_failed":
            identity = event.track_subscription_failed.participant_identity
            rparticipant = self._remote_participants[identity]
            error = event.track_subscription_failed.error
            self.emit(
                "track_subscription_failed",
                rparticipant,
                event.track_subscription_failed.track_sid,
                error,
            )
        elif which == "track_muted":
            identity = event.track_muted.participant_identity
            # TODO: pass participant identity
            participant = self._retrieve_participant(identity)
            assert isinstance(participant, Participant)
            publication = participant.track_publications[event.track_muted.track_sid]
            publication._info.muted = True
            if publication.track:
                publication.track._info.muted = True

            self.emit("track_muted", participant, publication)
        elif which == "track_unmuted":
            identity = event.track_unmuted.participant_identity
            # TODO: pass participant identity
            participant = self._retrieve_participant(identity)
            assert isinstance(participant, Participant)
            publication = participant.track_publications[event.track_unmuted.track_sid]
            publication._info.muted = False
            if publication.track:
                publication.track._info.muted = False

            self.emit("track_unmuted", participant, publication)
        elif which == "active_speakers_changed":
            speakers: list[Participant] = []
            # TODO: pass participant identity
            for identity in event.active_speakers_changed.participant_identities:
                participant = self._retrieve_participant(identity)
                assert isinstance(participant, Participant)
                speakers.append(participant)

            self.emit("active_speakers_changed", speakers)
        elif which == "room_metadata_changed":
            old_metadata = self.metadata
            self._info.metadata = event.room_metadata_changed.metadata
            self.emit("room_metadata_changed", old_metadata, self.metadata)
        elif which == "room_sid_changed":
            if not self._info.sid:
                self._first_sid_future.set_result(event.room_sid_changed.sid)
            self._info.sid = event.room_sid_changed.sid
            # This is an internal event, not exposed to users
        elif which == "participant_metadata_changed":
            identity = event.participant_metadata_changed.participant_identity
            # TODO: pass participant identity
            participant = self._retrieve_participant(identity)
            assert isinstance(participant, Participant)
            old_metadata = participant.metadata
            participant._info.metadata = event.participant_metadata_changed.metadata
            self.emit(
                "participant_metadata_changed",
                participant,
                old_metadata,
                participant.metadata,
            )
        elif which == "participant_name_changed":
            identity = event.participant_name_changed.participant_identity
            participant = self._retrieve_participant(identity)
            assert isinstance(participant, Participant)
            old_name = participant.name
            participant._info.name = event.participant_name_changed.name
            self.emit("participant_name_changed", participant, old_name, participant.name)
        elif which == "participant_attributes_changed":
            identity = event.participant_attributes_changed.participant_identity
            attributes = event.participant_attributes_changed.attributes
            changed_attributes = dict(
                (entry.key, entry.value)
                for entry in event.participant_attributes_changed.changed_attributes
            )
            participant = self._retrieve_participant(identity)
            assert isinstance(participant, Participant)
            participant._info.attributes.clear()
            participant._info.attributes.update((entry.key, entry.value) for entry in attributes)
            self.emit(
                "participant_attributes_changed",
                changed_attributes,
                participant,
            )
        elif which == "participant_encryption_status_changed":
            identity = event.participant_encryption_status_changed.participant_identity
            participant = self._retrieve_participant(identity)
            self.emit(
                "participant_encryption_status_changed",
                participant,
                event.participant_encryption_status_changed.is_encrypted,
            )
        elif which == "connection_quality_changed":
            identity = event.connection_quality_changed.participant_identity
            # TODO: pass participant identity
            participant = self._retrieve_participant(identity)
            self.emit(
                "connection_quality_changed",
                participant,
                event.connection_quality_changed.quality,
            )
        elif which == "transcription_received":
            transcription = event.transcription_received
            segments = [
                TranscriptionSegment(
                    id=s.id,
                    text=s.text,
                    final=s.final,
                    start_time=s.start_time,
                    end_time=s.end_time,
                    language=s.language,
                )
                for s in transcription.segments
            ]
            part = self._retrieve_participant(transcription.participant_identity)
            pub: TrackPublication | None = None
            if part:
                pub = part.track_publications.get(transcription.track_sid)
            self.emit("transcription_received", segments, part, pub)
        elif which == "data_packet_received":
            packet = event.data_packet_received
            which_val = packet.WhichOneof("value")
            if which_val == "user":
                owned_buffer_info = packet.user.data
                buffer_info = owned_buffer_info.data
                native_data = ctypes.cast(
                    buffer_info.data_ptr,
                    ctypes.POINTER(ctypes.c_byte * buffer_info.data_len),
                ).contents

                data = bytes(native_data)
                FfiHandle(owned_buffer_info.handle.id)
                rparticipant = cast(
                    RemoteParticipant,
                    self._retrieve_remote_participant(packet.participant_identity),
                )
                self.emit(
                    "data_received",
                    DataPacket(
                        data=data,
                        kind=packet.kind,
                        participant=rparticipant,
                        topic=packet.user.topic,
                    ),
                )
            elif which_val == "sip_dtmf":
                rparticipant = cast(
                    RemoteParticipant,
                    self._retrieve_remote_participant(packet.participant_identity),
                )
                self.emit(
                    "sip_dtmf_received",
                    SipDTMF(
                        code=packet.sip_dtmf.code,
                        digit=packet.sip_dtmf.digit,
                        participant=rparticipant,
                    ),
                )
        elif which == "e2ee_state_changed":
            identity = event.e2ee_state_changed.participant_identity
            e2ee_state = event.e2ee_state_changed.state
            # TODO: pass participant identity
            self.emit("e2ee_state_changed", self._retrieve_participant(identity), e2ee_state)
        elif which == "connection_state_changed":
            connection_state = event.connection_state_changed.state
            self._connection_state = connection_state
            self.emit("connection_state_changed", connection_state)
        elif which == "connected":
            self.emit("connected")
        elif which == "disconnected":
            self.emit("disconnected", event.disconnected.reason)
        elif which == "reconnecting":
            self.emit("reconnecting")
        elif which == "reconnected":
            self.emit("reconnected")
        elif which == "stream_header_received":
            self._handle_stream_header(
                event.stream_header_received.header,
                event.stream_header_received.participant_identity,
            )
        elif which == "stream_chunk_received":
            task = asyncio.create_task(self._handle_stream_chunk(event.stream_chunk_received.chunk))
            self._data_stream_tasks.add(task)
            task.add_done_callback(self._data_stream_tasks.discard)

        elif which == "stream_trailer_received":
            task = asyncio.create_task(
                self._handle_stream_trailer(event.stream_trailer_received.trailer)
            )
            self._data_stream_tasks.add(task)
            task.add_done_callback(self._data_stream_tasks.discard)

        elif which == "room_updated":
            self._info = event.room_updated
            self.emit("room_updated")

        elif which == "moved":
            self._info = event.moved
            self.emit("moved")

        elif which == "participants_updated":
            for info in event.participants_updated.participants:
                participant = self._retrieve_participant(info.identity)
                if participant:
                    participant._info = info

    def _handle_stream_header(
        self, header: proto_room.DataStream.Header, participant_identity: str
    ):
        stream_type = header.WhichOneof("content_header")
        if stream_type == "text_header":
            text_stream_handler = self._text_stream_handlers.get(header.topic)
            if text_stream_handler is None:
                logging.info(
                    "ignoring text stream with topic '%s', no callback attached",
                    header.topic,
                )
                return

            text_reader = TextStreamReader(header)
            self._text_stream_readers[header.stream_id] = text_reader
            text_stream_handler(text_reader, participant_identity)
        elif stream_type == "byte_header":
            byte_stream_handler = self._byte_stream_handlers.get(header.topic)
            if byte_stream_handler is None:
                logging.info(
                    "ignoring byte stream with topic '%s', no callback attached",
                    header.topic,
                )
                return

            byte_reader = ByteStreamReader(header)
            self._byte_stream_readers[header.stream_id] = byte_reader
            byte_stream_handler(byte_reader, participant_identity)
        else:
            logging.warning("received unknown header type, %s", stream_type)
        pass

    async def _handle_stream_chunk(self, chunk: proto_room.DataStream.Chunk):
        text_reader = self._text_stream_readers.get(chunk.stream_id)
        file_reader = self._byte_stream_readers.get(chunk.stream_id)

        if text_reader:
            await text_reader._on_chunk_update(chunk)
        elif file_reader:
            await file_reader._on_chunk_update(chunk)

    async def _handle_stream_trailer(self, trailer: proto_room.DataStream.Trailer):
        text_reader = self._text_stream_readers.get(trailer.stream_id)
        file_reader = self._byte_stream_readers.get(trailer.stream_id)

        if text_reader:
            await text_reader._on_stream_close(trailer)
            self._text_stream_readers.pop(trailer.stream_id)
        elif file_reader:
            await file_reader._on_stream_close(trailer)
            self._byte_stream_readers.pop(trailer.stream_id)

    async def _drain_rpc_invocation_tasks(self) -> None:
        if self._rpc_invocation_tasks:
            for task in self._rpc_invocation_tasks:
                task.cancel()
            await asyncio.gather(*self._rpc_invocation_tasks, return_exceptions=True)

    async def _drain_data_stream_tasks(self) -> None:
        if self._data_stream_tasks:
            for task in self._data_stream_tasks:
                task.cancel()
            await asyncio.gather(*self._data_stream_tasks, return_exceptions=True)

    def _retrieve_remote_participant(self, identity: str) -> Optional[RemoteParticipant]:
        """Retrieve a remote participant by identity"""
        return self._remote_participants.get(identity, None)

    def _retrieve_participant(self, identity: str) -> Optional[Participant]:
        """Retrieve a local or remote participant by identity"""
        if identity and identity == self.local_participant.identity:
            return self.local_participant

        return self._retrieve_remote_participant(identity)

    def _create_remote_participant(
        self, owned_info: proto_participant.OwnedParticipant
    ) -> RemoteParticipant:
        if owned_info.info.identity in self._remote_participants:
            raise Exception("participant already exists")

        participant = RemoteParticipant(owned_info)
        self._remote_participants[participant.identity] = participant
        return participant

    def __repr__(self) -> str:
        sid = "unknown"
        if self._first_sid_future.done():
            sid = self._first_sid_future.result()

        return f"rtc.Room(sid={sid}, name={self.name}, metadata={self.metadata}, connection_state={ConnectionState.Name(self._connection_state)})"



# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
from sysconfig import get_platform
from typing import Any, Dict

import setuptools  # type: ignore
import setuptools.command.build_py  # type: ignore
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel  # type: ignore

here = pathlib.Path(__file__).parent.resolve()
about: Dict[Any, Any] = {}
with open(os.path.join(here, "livekit", "rtc", "version.py"), "r") as f:
    exec(f.read(), about)


setuptools.setup(
    name="livekit",
    version=about["__version__"],
    description="Python Real-time SDK for LiveKit",
    long_description=(here / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://github.com/livekit/python-sdks",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Video",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords=["webrtc", "realtime", "audio", "video", "livekit"],
    license="Apache-2.0",
    packages=setuptools.find_namespace_packages(include=["livekit.*"]),
    python_requires=">=3.9.0",
    install_requires=["protobuf>=4.25.0", "types-protobuf>=3", "aiofiles>=24", "numpy>=1.26", "av>=11.0.0"],
    package_data={
        "livekit.rtc": ["_proto/*.py", "py.typed", "*.pyi", "**/*.pyi"],
        "livekit.rtc.resources": [
            "*.so",
            "*.dylib",
            "*.dll",
            "LICENSE.md",
            "*.h",
            "jupyter-html/index.html",
        ],
    },
    project_urls={
        "Documentation": "https://docs.livekit.io",
        "Website": "https://livekit.io/",
        "Source": "https://github.com/livekit/python-sdks/",
    },
)



# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, List, Union
from ._ffi_client import FfiHandle, FfiClient
from ._proto import ffi_pb2 as proto_ffi
from ._proto import track_pb2 as proto_track
from ._proto import stats_pb2 as proto_stats

if TYPE_CHECKING:
    from .audio_source import AudioSource
    from .video_source import VideoSource


class Track:
    def __init__(self, owned_info: proto_track.OwnedTrack):
        self._info = owned_info.info
        self._ffi_handle = FfiHandle(owned_info.handle.id)

    @property
    def sid(self) -> str:
        return self._info.sid

    @property
    def name(self) -> str:
        return self._info.name

    @property
    def kind(self) -> proto_track.TrackKind.ValueType:
        return self._info.kind

    @property
    def stream_state(self) -> proto_track.StreamState.ValueType:
        return self._info.stream_state

    @property
    def muted(self) -> bool:
        return self._info.muted

    async def get_stats(self) -> List[proto_stats.RtcStats]:
        req = proto_ffi.FfiRequest()
        req.get_stats.track_handle = self._ffi_handle.handle

        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            cb: proto_ffi.FfiEvent = await queue.wait_for(
                lambda e: e.get_stats.async_id == resp.get_stats.async_id
            )
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

        if cb.get_stats.error:
            raise Exception(cb.get_stats.error)

        return list(cb.get_stats.stats)


class LocalAudioTrack(Track):
    def __init__(self, info: proto_track.OwnedTrack):
        super().__init__(info)

    @staticmethod
    def create_audio_track(name: str, source: "AudioSource") -> "LocalAudioTrack":
        req = proto_ffi.FfiRequest()
        req.create_audio_track.name = name
        req.create_audio_track.source_handle = source._ffi_handle.handle

        resp = FfiClient.instance.request(req)
        return LocalAudioTrack(resp.create_audio_track.track)

    def mute(self):
        req = proto_ffi.FfiRequest()
        req.local_track_mute.track_handle = self._ffi_handle.handle
        req.local_track_mute.mute = True
        FfiClient.instance.request(req)
        self._info.muted = True

    def unmute(self):
        req = proto_ffi.FfiRequest()
        req.local_track_mute.track_handle = self._ffi_handle.handle
        req.local_track_mute.mute = False
        FfiClient.instance.request(req)
        self._info.muted = False

    def __repr__(self) -> str:
        return f"rtc.LocalAudioTrack(sid={self.sid}, name={self.name})"


class LocalVideoTrack(Track):
    def __init__(self, info: proto_track.OwnedTrack):
        super().__init__(info)

    @staticmethod
    def create_video_track(name: str, source: "VideoSource") -> "LocalVideoTrack":
        req = proto_ffi.FfiRequest()
        req.create_video_track.name = name
        req.create_video_track.source_handle = source._ffi_handle.handle

        resp = FfiClient.instance.request(req)
        return LocalVideoTrack(resp.create_video_track.track)

    def mute(self):
        req = proto_ffi.FfiRequest()
        req.local_track_mute.track_handle = self._ffi_handle.handle
        req.local_track_mute.mute = True
        FfiClient.instance.request(req)
        self._info.muted = True

    def unmute(self):
        req = proto_ffi.FfiRequest()
        req.local_track_mute.track_handle = self._ffi_handle.handle
        req.local_track_mute.mute = False
        FfiClient.instance.request(req)
        self._info.muted = False

    def __repr__(self) -> str:
        return f"rtc.LocalVideoTrack(sid={self.sid}, name={self.name})"


class RemoteAudioTrack(Track):
    def __init__(self, info: proto_track.OwnedTrack):
        super().__init__(info)

    def __repr__(self) -> str:
        return f"rtc.RemoteAudioTrack(sid={self.sid}, name={self.name})"


class RemoteVideoTrack(Track):
    def __init__(self, info: proto_track.OwnedTrack):
        super().__init__(info)

    def __repr__(self) -> str:
        return f"rtc.RemoteVideoTrack(sid={self.sid}, name={self.name})"


LocalTrack = Union[LocalVideoTrack, LocalAudioTrack]
RemoteTrack = Union[RemoteVideoTrack, RemoteAudioTrack]
AudioTrack = Union[LocalAudioTrack, RemoteAudioTrack]
VideoTrack = Union[LocalVideoTrack, RemoteVideoTrack]
